import multiprocessing
import traceback
import pickle
from typing import Dict, List
from aworld.replay_buffer.base import Storage, DataRow
from aworld.replay_buffer.query_filter import QueryCondition, QueryFilter
from aworld.logs.util import logger


class MultiProcMemoryStorage(Storage):

    """
    Memory storage for multi-process.
    """

    def __init__(self,
                 data_dict: Dict[str, str],
                 fifo_queue: List[str],
                 lock: multiprocessing.Lock,
                 max_capacity: int = 10000):
        self._data: Dict[str, str] = data_dict
        self._fifo_queue = fifo_queue
        self._max_capacity = max_capacity
        self._lock = lock

    def _save_to_shared_memory(self, data, task_id):
        serialized_data = pickle.dumps(data)
        try:
            if task_id not in self._data or not self._data[task_id]:
                shm = multiprocessing.shared_memory.SharedMemory(
                    create=True, size=len(serialized_data))
                shm.buf[:len(serialized_data)] = serialized_data
                self._data[task_id] = shm.name
                shm.close()
                return
            shm = multiprocessing.shared_memory.SharedMemory(
                name=self._data[task_id], create=False)
            if len(serialized_data) > shm.size:
                shm.close()
                shm.unlink()
                shm = multiprocessing.shared_memory.SharedMemory(
                    create=True, size=len(serialized_data))
                shm.buf[:len(serialized_data)] = serialized_data
                self._data[task_id] = shm.name
            else:
                shm.buf[:len(serialized_data)] = serialized_data
        except FileNotFoundError:
            shm = multiprocessing.shared_memory.SharedMemory(
                create=True, size=len(serialized_data))
            shm.buf[:len(serialized_data)] = serialized_data
            self._data[task_id] = shm.name
        shm.close()

    def _load_from_shared_memory(self, task_id):
        try:
            if task_id not in self._data or not self._data[task_id]:
                return []
            try:
                multiprocessing.shared_memory.SharedMemory(
                    name=self._data[task_id], create=False)
            except FileNotFoundError:
                return []
            shm = multiprocessing.shared_memory.SharedMemory(
                name=self._data[task_id])
            data = pickle.loads(shm.buf.tobytes())
            shm.close()
            return data
        except Exception as e:
            stack_trace = traceback.format_exc()
            logger.error(
                f"_load_from_shared_memory error: {e}\nStack trace:\n{stack_trace}")
            return []

    def _delete_from_shared_memory(self, task_id):
        try:
            if task_id not in self._data or not self._data[task_id]:
                return
            shm = multiprocessing.shared_memory.SharedMemory(
                name=self._data[task_id])
            shm.close()
            shm.unlink()
            del self._data[task_id]
        except FileNotFoundError:
            pass

    def add(self, data: DataRow):
        if not data:
            raise ValueError("Data is required")
        if not data.exp_meta:
            raise ValueError("exp_meta is required")

        with self._lock:
            current_size = sum(len(self._load_from_shared_memory(task_id))
                               for task_id in self._data.keys())
            while current_size >= self._max_capacity and self._fifo_queue:
                oldest_task_id = self._fifo_queue.pop(0)
                if oldest_task_id in self._data.keys():
                    current_size -= len(self._load_from_shared_memory(oldest_task_id))
                    self._delete_from_shared_memory(oldest_task_id)

            task_id = data.exp_meta.task_id
            existing_data = self._load_from_shared_memory(task_id)
            existing_data.append(data)
            self._save_to_shared_memory(existing_data, task_id)
            self._fifo_queue.append(task_id)

    def add_batch(self, data_batch: List[DataRow]):
        with self._lock:
            for data in data_batch:
                self.add(data)

    def size(self, query_condition: QueryCondition = None) -> int:
        with self._lock:
            return len(self._get_all_without_lock(query_condition))

    def get_paginated(self, page: int, page_size: int, query_condition: QueryCondition = None) -> List[DataRow]:
        with self._lock:
            if page < 1:
                raise ValueError("Page must be greater than 0")
            if page_size < 1:
                raise ValueError("Page size must be greater than 0")
            all_data = self._get_all_without_lock(query_condition)
            start_index = (page - 1) * page_size
            end_index = start_index + page_size
            return all_data[start_index:end_index]

    def _get_all_without_lock(self, query_condition: QueryCondition = None) -> List[DataRow]:
        all_data = []
        query_filter = None
        if query_condition:
            query_filter = QueryFilter(query_condition)
        for task_id in self._data.keys():
            local_data = self._load_from_shared_memory(task_id)
            if query_filter:
                all_data.extend(query_filter.filter(local_data))
            else:
                all_data.extend(local_data)
        return all_data

    def get_all(self, query_condition: QueryCondition = None) -> List[DataRow]:
        with self._lock:
            return self._get_all_without_lock(query_condition)

    def get_by_task_id(self, task_id: str) -> List[DataRow]:
        with self._lock:
            if task_id in self._data.keys():
                return self._load_from_shared_memory(task_id)

    def get_bacth_by_task_ids(self, task_ids: List[str]) -> Dict[str, List[DataRow]]:
        with self._lock:
            result = {}
            for task_id in task_ids:
                if task_id in self._data.keys():
                    result[task_id] = self._load_from_shared_memory(task_id)
            return result



import datetime
import hashlib
import uuid
from abc import ABC, abstractmethod
from typing import Optional, Any

from pydantic import BaseModel, Field


class MemoryItem(BaseModel):
    id: str = Field(description="id")
    content: Any = Field(description="content")
    hash: str = Field(description="content hash")
    created_at: Optional[str] = Field(None, description="created at")
    updated_at: Optional[str] = Field(None, description="updated at")
    metadata: dict = Field(
        description="metadata, use to store additional information, such as user_id, agent_id, run_id, task_id, etc.")
    tags: list[str] = Field(description="tags")
    version: int = Field(description="version")

    def __init__(self, **data):
        # Set default values for optional fields
        if "id" not in data:
            data["id"] = str(uuid.uuid4())
        if "hash" not in data:
            data["hash"] = hashlib.md5((data.get("content", "") or "").encode()).hexdigest()
        if "created_at" not in data:
            data["created_at"] = datetime.datetime.now().isoformat()
        if "updated_at" not in data:
            data["updated_at"] = data["created_at"]
        if "metadata" not in data:
            data["metadata"] = {}
        if "tags" not in data:
            data["tags"] = []
        if "version" not in data:
            data["version"] = 1

        super().__init__(**data)

    @classmethod
    def from_dict(cls, data: dict) -> "MemoryItem":
        """Create a MemoryItem instance from a dictionary.

        Args:
            data (dict): A dictionary containing the memory item data.

        Returns:
            MemoryItem: An instance of MemoryItem.
        """
        return cls(**data)


class MemoryStore(ABC):
    @abstractmethod
    def add(self, memory_item: MemoryItem):
        pass

    @abstractmethod
    def get(self, memory_id) -> Optional[MemoryItem]:
        pass

    @abstractmethod
    def get_first(self) -> Optional[MemoryItem]:
        pass

    @abstractmethod
    def total_rounds(self) -> int:
        pass

    @abstractmethod
    def get_all(self) -> list[MemoryItem]:
        pass

    @abstractmethod
    def get_last_n(self, last_rounds) -> list[MemoryItem]:
        pass

    @abstractmethod
    def update(self, memory_item: MemoryItem):
        pass

    @abstractmethod
    def delete(self, memory_id):
        pass

    @abstractmethod
    def retrieve(self, query, filters: dict) -> list[MemoryItem]:
        pass

    @abstractmethod
    def history(self, memory_id) -> list[MemoryItem]:
        pass


class MemoryBase(ABC):
    @abstractmethod
    def retrieve(self, query, filters: dict) -> list[MemoryItem]:
        """Retrieve a memory by ID.

        Args:
            query (str): query.
            filters (dict): filters.

        Returns:
            list: Retrieved memory.
        """

    @abstractmethod
    def get(self, memory_id) -> Optional[MemoryItem]:
        """Get item in memory by ID.

        Args:
            memory_id (str): ID of the memory to retrieve.

        Returns:
            dict: Retrieved memory.
        """

    @abstractmethod
    def get_all(self) -> list[MemoryItem]:
        """List all items in memory store.

        Returns:
            list: List of all memories.
        """

    @abstractmethod
    def get_last_n(self, last_rounds) -> list[MemoryItem]:
        """get last_rounds memories.

        Returns:
            list: List of latest memories.
        """

    @abstractmethod
    def add(self, memory_item: MemoryItem):
        """Add memory in the memory store.


        Args:
            memory_item (MemoryItem): memory item.
        """

    @abstractmethod
    def update(self, memory_item: MemoryItem):
        """Update a memory by ID.

        Args:
            memory_item (MemoryItem): memory item.

        Returns:
            dict: Updated memory.
        """

    @abstractmethod
    def delete(self, memory_id):
        """Delete a memory by ID.

        Args:
            memory_id (str): ID of the memory to delete.
        """

    @abstractmethod
    def history(self, memory_id) -> list[MemoryItem]:
        """Get the history of changes for a memory by ID.

        Args:
            memory_id (str): ID of the memory to get history for.

        Returns:
            list: List of changes for the memory.
        """



import json
from dataclasses import dataclass

from aworld.output import (
    MessageOutput,
    AworldUI,
    Output,
)
from aworld.output.base import StepOutput, ToolResultOutput
from aworld.output.utils import consume_content
import re


tool_call_template = """

**call {function_name}**

```tool_call_arguments
{function_arguments}
```
        
```tool_call_result
{function_result}
```

"""

step_loading_template = """
```loading
{data}
```
"""


@dataclass
class MarkdownAworldUI(AworldUI):

    def __init__(self, **kwargs):
        """
        Initialize MarkdownAworldUI
        Args:"""
        super().__init__(**kwargs)

    async def message_output(self, __output__: MessageOutput):

        items = []

        async def __log_item(item):
            items.append(item)

        await consume_content(__output__.reasoning, __log_item)
        await consume_content(__output__.response, __log_item)

        return items

    async def tool_result(self, output: ToolResultOutput):
        """
        tool_result
        """
        fn_args = await self.json_parse(output.origin_tool_call.function.arguments)
        fn_args = re.sub(r"```", "````", fn_args, flags=re.MULTILINE)
        fn_results = await self.json_parse(output.data)
        fn_results = re.sub(r"```", "````", fn_results, flags=re.MULTILINE)
        tool_data = tool_call_template.format(
            function_name=output.origin_tool_call.function.name,
            function_arguments=fn_args,
            function_result=fn_results,
        )
        return tool_data

    async def json_parse(self, json_str):
        try:
            function_result = json.dumps(
                json.loads(json_str), indent=2, ensure_ascii=False
            )
        except Exception:
            function_result = json_str
        return function_result

    async def step(self, output: StepOutput):
        emptyLine = "\n\n----\n\n"
        if output.status == "START":
            return f"\n### {output.name} 🛫START"
        elif output.status == "FINISHED":
            return f"\n{output.name} 🛬FINISHED {emptyLine}"
        elif output.status == "FAILED":
            return f"\n{output.name} 💥FAILED {emptyLine}"
        else:
            return f"\n{output.name} ❓❓❓UNKNOWN#{output.status} {emptyLine}"

    async def custom_output(self, output: Output):
        return output.data



import logging
import asyncio
import uuid
from typing import Dict, List, Any, Optional

from aworld.sandbox.common import BaseSandbox
from aworld.sandbox.api.kubernetes.sandbox_api import KubernetesSandboxApi
from aworld.sandbox.models import SandboxStatus, SandboxEnvType, SandboxInfo
from aworld.sandbox.run.mcp_servers import McpServers


class KubernetesSandbox(BaseSandbox, KubernetesSandboxApi):
    """
    Kubernetes sandbox implementation that runs in a Kubernetes cluster.
    """

    def __init__(
            self,
            sandbox_id: Optional[str] = None,
            metadata: Optional[Dict[str, str]] = None,
            timeout: Optional[int] = None,
            mcp_servers: Optional[List[str]] = None,
            mcp_config: Optional[Any] = None,
            **kwargs
    ):
        """
        Initialize a new KubernetesSandbox instance.
        
        Args:
            sandbox_id: Unique identifier for the sandbox. If None, one will be generated.
            metadata: Additional metadata for the sandbox.
            timeout: Timeout for sandbox operations.
            mcp_servers: List of MCP servers to use.
            mcp_config: Configuration for MCP servers.
            **kwargs: Additional parameters for specific sandbox types.
        """
        super().__init__(
            sandbox_id=sandbox_id,
            env_type=SandboxEnvType.K8S,
            metadata=metadata,
            timeout=timeout,
            mcp_servers=mcp_servers,
            mcp_config=mcp_config
        )

        if sandbox_id:
            if not self._metadata:
                return self
            else:
                raise ValueError("sandbox_id is not exist")

        # Initialize properties
        self._status = SandboxStatus.INIT
        self._timeout = timeout or self.default_sandbox_timeout
        self._metadata = metadata or {}
        self._env_type = SandboxEnvType.K8S
        self._mcp_servers = mcp_servers
        self._mcp_config = mcp_config
        
        # Ensure sandbox_id has a value in all cases
        self._sandbox_id = sandbox_id or str(uuid.uuid4())

        # If no sandbox_id provided, create a new sandbox
        if not sandbox_id:
            response = self._create_sandbox(
                env_type=self._env_type,
                env_config=None,
                mcp_servers=mcp_servers,
                mcp_config=mcp_config,
            )
            
            if not response:
                self._status = SandboxStatus.ERROR
                # If creation fails, keep the generated UUID as the ID
                logging.warning(f"Failed to create K8s sandbox, using generated ID: {self._sandbox_id}")
            else:
                self._sandbox_id = response.sandbox_id
                self._status = SandboxStatus.RUNNING
                self._metadata = {
                    "pod_name": getattr(response, 'pod_name', None),
                    "service_name": getattr(response, 'service_name', None),
                    "status": getattr(response, 'status', None),
                    "cluster_ip": getattr(response, 'cluster_ip', None),
                    "host": getattr(response, 'host', None),
                    "mcp_config": getattr(response, 'mcp_config', None),
                    "env_type": getattr(response, 'env_type', None),
                }
                self._mcp_config = getattr(response, 'mcp_config', None)
            
        # Initialize McpServers
        self._mcpservers = McpServers(
            mcp_servers,
            self._mcp_config,
        )

    async def remove(self) -> None:
        """
        Remove sandbox.
        """
        await self._remove_sandbox(
            sandbox_id=self.sandbox_id,
            metadata=self._metadata,
            env_type=self._env_type
        )
        
    async def cleanup(self) -> None:
        """
        Clean up Sandbox resources, including MCP server connections
        """
        try:
            if hasattr(self, '_mcpservers') and self._mcpservers:
                await self._mcpservers.cleanup()
                logging.info(f"Cleaned up MCP servers for sandbox {self.sandbox_id}")
        except Exception as e:
            logging.warning(f"Failed to cleanup MCP servers: {e}")
        
        # Call the original remove method
        try:
            await self.remove()
        except Exception as e:
            logging.warning(f"Failed to remove sandbox: {e}")
            
    def __del__(self):
        """
        Ensure resources are cleaned up when the object is garbage collected
        """
        try:
            # Handle the case where an event loop already exists
            try:
                loop = asyncio.get_running_loop()
                logging.warning("Cannot clean up sandbox in __del__ when event loop is already running")
                return
            except RuntimeError:
                # No running event loop, create a new one
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                loop.run_until_complete(self.cleanup())
                loop.close()
        except Exception as e:
            logging.warning(f"Failed to cleanup sandbox resources during garbage collection: {e}") 


# coding: utf-8
# Copyright (c) 2025 inclusionAI.
from typing import Any, List

from aworld.core.common import ActionResult, Observation

DEFAULT_VIRTUAL_ENV_ID = "env_0"


def build_observation(observer: str,
                      ability: str,
                      container_id: str = None,
                      content: Any = None,
                      dom_tree: Any = None,
                      action_result: List[ActionResult] = [],
                      image: str = '',
                      images: List[str] = [],
                      **kwargs):
    return Observation(container_id=container_id if container_id else DEFAULT_VIRTUAL_ENV_ID,
                       observer=observer,
                       ability=ability,
                       content=content,
                       action_result=action_result,
                       dom_tree=dom_tree,
                       image=image,
                       images=images,
                       info=kwargs)



"""
Download MCP Server

This module provides MCP server functionality for downloading files from URLs.
It handles various download scenarios with proper validation, error handling,
and progress tracking.

Key features:
- File downloading from HTTP/HTTPS URLs
- Download progress tracking
- File validation
- Safe file saving

Main functions:
- mcpdownload: Downloads files from URLs to local filesystem
"""

import os
import sys
import traceback
import urllib.parse
from pathlib import Path
from typing import List, Optional

import requests
from dotenv import load_dotenv
from mcp.server.fastmcp import FastMCP
from pydantic import BaseModel, Field

from aworld.logs.util import logger

mcp = FastMCP("download-server")


class DownloadResult(BaseModel):
    """Download result model with file information"""

    file_path: str
    file_name: str
    file_size: int
    content_type: Optional[str] = None
    success: bool
    error: Optional[str] = None


class DownloadResults(BaseModel):
    """Download results model for multiple files"""

    results: List[DownloadResult]
    success_count: int
    failed_count: int


@mcp.tool(description="Download files from URLs and save to the local filesystem.")
def mcpdownloadfiles(
    urls: List[str] = Field(
        ..., description="The URLs of the files to download. Must be a list of URLs."
    ),
    output_dir: str = Field(
        "/tmp/mcp_downloads",
        description="Directory to save the downloaded files (default: /tmp/mcp_downloads).",
    ),
    timeout: int = Field(60, description="Download timeout in seconds (default: 60)."),
) -> str:
    """Download files from URLs and save to the local filesystem.

    Args:
        urls: The URLs of the files to download, must be a list of URLs
        output_dir: Directory to save the downloaded files
        timeout: Download timeout in seconds

    Returns:
        JSON string with download results information
    """
    results = []
    success_count = 0
    failed_count = 0

    for single_url in urls:
        result_json = _download_single_file(single_url, output_dir, "", timeout)
        result = DownloadResult.model_validate_json(result_json)
        results.append(result)

        if result.success:
            success_count += 1
        else:
            failed_count += 1

    batch_results = DownloadResults(
        results=results, success_count=success_count, failed_count=failed_count
    )

    return batch_results.model_dump_json()


def _download_single_file(
    url: str, output_dir: str, filename: str, timeout: int
) -> str:
    """Download a single file from URL and save it to the local filesystem."""
    try:
        # Validate URL
        if not url.startswith(("http://", "https://")):
            raise ValueError(
                "Invalid URL format. URL must start with http:// or https://"
            )

        # Create output directory if it doesn't exist
        output_path = Path(output_dir)
        output_path.mkdir(parents=True, exist_ok=True)

        # Determine filename if not provided
        if not filename:
            filename = os.path.basename(urllib.parse.urlparse(url).path)
            if not filename:
                filename = "downloaded_file"

        # Full path to save the file
        file_path = os.path.join(output_path, filename)

        logger.info(f"Downloading file from {url} to {file_path}")
        # Download the file with progress tracking
        headers = {
            "User-Agent": (
                "Mozilla/5.0 (Windows NT 10.0; Win64; x64) "
                "AWorld/1.0 (https://github.com/inclusionAI/AWorld; qintong.wqt@antgroup.com) "
                "Python/requests "
            ),
            "Accept": "text/html,application/xhtml+xml,application/xml,application/pdf;q=0.9,image/webp,*/*;q=0.8",
            "Accept-Language": "en-US,en;q=0.5",
            "Accept-Encoding": "gzip, deflate, br",
            "Connection": "keep-alive",
        }

        response = requests.get(url, headers=headers, stream=True, timeout=timeout)
        response.raise_for_status()

        # Get content type and size
        content_type = response.headers.get("Content-Type")

        # Save the file
        with open(file_path, "wb") as f:
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:
                    f.write(chunk)

        # Get actual file size
        actual_size = os.path.getsize(file_path)

        logger.info(f"File downloaded successfully to {file_path}")

        # Create result
        result = DownloadResult(
            file_path=file_path,
            file_name=filename,
            file_size=actual_size,
            content_type=content_type,
            success=True,
            error=None,
        )

        return result.model_dump_json()

    except Exception as e:
        error_msg = str(e)
        logger.error(f"Download error: {traceback.format_exc()}")

        result = DownloadResult(
            file_path="",
            file_name="",
            file_size=0,
            content_type=None,
            success=False,
            error=error_msg,
        )

        return result.model_dump_json()


def main():
    load_dotenv()

    print("Starting Download MCP Server...", file=sys.stderr)
    mcp.run(transport="stdio")


# Make the module callable
def __call__():
    """
    Make the module callable for uvx.
    This function is called when the module is executed directly.
    """
    main()


sys.modules[__name__].__call__ = __call__

# Run the server when the script is executed directly
if __name__ == "__main__":
    main()



from aworld.trace.base import Propagator, Carrier, TraceContext
from aworld.trace.baggage import BaggageContext
from aworld.logs.util import logger
from aworld.trace.base import AttributeValueType


class SofaTracerBaggagePropagator(Propagator):
    """
    Sofa tracer baggage propagator.
    """

    _TRACE_ID_HEDER_NAMES = ["SOFA-TraceId", "sofaTraceId"]
    _SPAN_ID_HEDER_NAMES = ["SOFA-RpcId", "sofaRpcId"]
    _PEN_ATTRS_HEDER_NAME = "sofaPenAttrs"
    _SYS_PEN_ATTRS_HEDER_NAME = "sysPenAttrs"

    _TRACE_ID_BAGGAGE_KEY = "attributes.sofa.traceid"
    _SPAN_ID_BAGGAGE_KEY = "attributes.sofa.rpcid"
    _PEN_ATTRS_BAGGAGE_KEY = "attributes.sofa.penattrs"
    _SYS_PEN_ATTRS_BAGGAGE_KEY = "attributes.sofa.syspenattrs"

    def extract(self, carrier: Carrier):
        """
        Extract trace context from carrier.
        Args:
            carrier: The carrier to extract trace context from.
        Returns:
            A dict of trace context.
        """
        trace_id = None
        span_id = None
        for name in self._TRACE_ID_HEDER_NAMES:
            trace_id = self._get_value(carrier, name)
            if trace_id:
                break
        for name in self._SPAN_ID_HEDER_NAMES:
            span_id = self._get_value(carrier, name)
            if span_id:
                break
        pen_attrs = self._get_value(carrier, self._PEN_ATTRS_HEDER_NAME)
        sys_pen_attrs = self._get_value(
            carrier, self._SYS_PEN_ATTRS_HEDER_NAME)

        logger.info(
            f"extract trace_id: {trace_id}, span_id: {span_id}, pen_attrs: {pen_attrs}, sys_pen_attrs: {sys_pen_attrs}")
        if trace_id and span_id:
            BaggageContext.set_baggage(self._TRACE_ID_BAGGAGE_KEY, trace_id)
            span_id = span_id + ".1"
            BaggageContext.set_baggage(self._SPAN_ID_BAGGAGE_KEY, span_id)
            if pen_attrs:
                BaggageContext.set_baggage(
                    self._PEN_ATTRS_BAGGAGE_KEY, pen_attrs)
            if sys_pen_attrs:
                BaggageContext.set_baggage(
                    self._SYS_PEN_ATTRS_BAGGAGE_KEY, sys_pen_attrs)

    def inject(self, trace_context: TraceContext, carrier: Carrier):
        """
        Inject trace context to carrier.
        Args:
            trace_context: The trace context to inject.
            carrier: The carrier to inject trace context to.
        """
        baggage = BaggageContext.get_baggage()

        if baggage:
            trace_id = baggage.get(self._TRACE_ID_BAGGAGE_KEY)
            span_id = baggage.get(self._SPAN_ID_BAGGAGE_KEY)
            if trace_id and span_id:
                carrier.set(self._TRACE_ID_HEDER_NAMES[0], trace_id)
                carrier.set(self._SPAN_ID_HEDER_NAMES[0], span_id)

            pen_attrs_dict = {}
            for key, value in baggage.items():
                if key == self._TRACE_ID_BAGGAGE_KEY or key == self._SPAN_ID_BAGGAGE_KEY:
                    continue
                if key == self._PEN_ATTRS_BAGGAGE_KEY and value:
                    pen_attrs_dict.update(dict(item.split("=")
                                          for item in value.split("&")))
                    continue
                if key == self._SYS_PEN_ATTRS_BAGGAGE_KEY and value:
                    carrier.set(self._SYS_PEN_ATTRS_HEDER_NAME, value)
                    continue

                # other baggage items will be injected to sofaPenAttrs
                pen_attrs_dict.update({key: value})

            if pen_attrs_dict:
                pen_attrs = "&".join(f"{key}={value}"
                                     for key, value in pen_attrs_dict.items())
                carrier.set(self._PEN_ATTRS_HEDER_NAME, pen_attrs)


class SofaSpanHelper:
    """
    Sofa span helper.
    """

    @staticmethod
    def set_sofa_context_to_attr(span_attributes: dict[str, AttributeValueType]):
        """
        Set sofa context to span attributes.
        Args:
            span_attributes: The span attributes to set sofa context to.
        """
        baggage = BaggageContext.get_baggage()
        if baggage:
            trace_id = baggage.get(
                SofaTracerBaggagePropagator._TRACE_ID_BAGGAGE_KEY)
            span_id = baggage.get(
                SofaTracerBaggagePropagator._SPAN_ID_BAGGAGE_KEY)
            if trace_id and span_id:
                span_attributes.update({
                    SofaTracerBaggagePropagator._TRACE_ID_BAGGAGE_KEY: trace_id,
                    SofaTracerBaggagePropagator._SPAN_ID_BAGGAGE_KEY: span_id
                })
            pen_attrs = baggage.get(
                SofaTracerBaggagePropagator._PEN_ATTRS_BAGGAGE_KEY)
            if pen_attrs:
                span_attributes.update({
                    SofaTracerBaggagePropagator._PEN_ATTRS_BAGGAGE_KEY: pen_attrs
                })
            sys_pen_attrs = baggage.get(
                SofaTracerBaggagePropagator._SYS_PEN_ATTRS_BAGGAGE_KEY)
            if sys_pen_attrs:
                span_attributes.update({
                    SofaTracerBaggagePropagator._SYS_PEN_ATTRS_BAGGAGE_KEY: sys_pen_attrs
                })



# coding: utf-8
# Copyright (c) 2025 inclusionAI.
from typing import Dict, Any, List, Callable

from aworld.core.context.base import Context
from aworld.core.event.event_bus import InMemoryEventbus
from aworld.core.event.base import Constants, Message


class EventManager:
    """The event manager is now used to build an event bus instance and store the messages recently."""

    def __init__(self, **kwargs):
        # use conf to build event bus instance
        self.event_bus = InMemoryEventbus()
        self.context = Context.instance()
        # Record events in memory for re-consume.
        self.messages: Dict[str, List[Message]] = {'None': []}
        self.max_len = kwargs.get('max_len', 1000)

    async def emit(
            self,
            data: Any,
            sender: str,
            receiver: str = None,
            topic: str = None,
            session_id: str = None,
            event_type: str = Constants.TASK
    ):
        """Send data to the event bus.

        Args:
            data: Message payload.
            sender: The sender name of the message.
            receiver: The receiver name of the message.
            topic: The topic to which the message belongs.
            session_id: Special session id.
            event_type: Event type.
        """
        event = Message(
            payload=data,
            session_id=session_id if session_id else self.context.session_id,
            sender=sender,
            receiver=receiver,
            topic=topic,
            category=event_type,
        )
        return await self.emit_message(event)

    async def emit_message(self, event: Message):
        """Send the message to the event bus."""
        key = event.key()
        if key not in self.messages:
            self.messages[key] = []
        self.messages[key].append(event)
        if len(self.messages) > self.max_len:
            self.messages = self.messages[-self.max_len:]

        await self.event_bus.publish(event)
        return True

    async def consume(self, nowait: bool = False):
        if nowait:
            return await self.event_bus.consume_nowait()
        return await self.event_bus.consume()

    async def register(self, event_type: str, topic: str, handler: Callable[..., Any], **kwargs):
        await self.event_bus.subscribe(event_type, topic, handler, **kwargs)

    async def unregister(self, event_type: str, topic: str, handler: Callable[..., Any], **kwargs):
        await self.event_bus.unsubscribe(event_type, topic, handler, **kwargs)

    async def register_transformer(self, event_type: str, topic: str, handler: Callable[..., Any], **kwargs):
        await self.event_bus.subscribe(event_type, topic, handler, transformer=True, **kwargs)

    async def unregister_transformer(self, event_type: str, topic: str, handler: Callable[..., Any], **kwargs):
        await self.event_bus.unsubscribe(event_type, topic, handler, transformer=True, **kwargs)

    def messages_by_key(self, key: str) -> List[Message]:
        return self.messages.get(key, [])

    def messages_by_sender(self, sender: str, key: str):
        results = []
        for res in self.messages.get(key, []):
            if res.sender == sender:
                results.append(res)
        return results

    def messages_by_topic(self, topic: str, key: str):
        results = []
        for res in self.messages.get(key, []):
            if res.topic == topic:
                results.append(res)
        return results

    def session_messages(self, session_id: str) -> List[Message]:
        return [m for k, msg in self.messages.items() for m in msg if m.session_id == session_id]

    @staticmethod
    def mark_valid(messages: List[Message]):
        for msg in messages:
            msg.is_valid = True

    @staticmethod
    def mark_invalid(messages: List[Message]):
        for msg in messages:
            msg.is_valid = False

    def clear_messages(self):
        self.messages = []



import logging
from typing import List, Dict, Any
import json
import os
from contextlib import AsyncExitStack
import traceback

import requests
from mcp.types import CallToolResult

from aworld.core.common import ActionResult

from aworld.logs.util import logger
from aworld.mcp_client.server import MCPServer, MCPServerSse, MCPServerStdio
from aworld.utils.common import find_file

MCP_SERVERS_CONFIG = {}

async def run(mcp_servers: list[MCPServer]) -> List[Dict[str, Any]]:
    openai_tools = []
    for i, server in enumerate(mcp_servers):
        try:
            tools = await server.list_tools()
            for tool in tools:
                required = []
                properties = {}
                if tool.inputSchema and tool.inputSchema.get("properties"):
                    required = tool.inputSchema.get("required", [])
                    _properties = tool.inputSchema["properties"]
                    for param_name, param_info in _properties.items():
                        param_type = param_info.get("type") if param_info.get("type") != "str" and param_info.get("type") is not None else "string"
                        param_desc = param_info.get("description", "")
                        if param_type == "array":
                            # Handle array type parameters
                            items_info = param_info.get("items", {})
                            item_type = items_info.get("type", "string")

                            # Process nested array type parameters
                            if item_type == "array":
                                nested_items = items_info.get("items", {})
                                nested_type = nested_items.get("type", "string")

                                # If the nested type is an object
                                if nested_type == "object":
                                    properties[param_name] = {
                                        "description": param_desc,
                                        "type": param_type,
                                        "items": {
                                            "type": item_type,
                                            "items": {
                                                "type": nested_type,
                                                "properties": nested_items.get("properties", {}),
                                                "required": nested_items.get("required", [])
                                            }
                                        }
                                    }
                                else:
                                    properties[param_name] = {
                                        "description": param_desc,
                                        "type": param_type,
                                        "items": {
                                            "type": item_type,
                                            "items": {
                                                "type": nested_type
                                            }
                                        }
                                    }
                            # Process object type cases
                            elif item_type == "object":
                                properties[param_name] = {
                                    "description": param_desc,
                                    "type": param_type,
                                    "items": {
                                        "type": item_type,
                                        "properties": items_info.get("properties", {}),
                                        "required": items_info.get("required", [])
                                    }
                                }
                            # Process basic type cases
                            else:
                                if item_type == "str":
                                    item_type = "string"
                                properties[param_name] = {
                                    "description": param_desc,
                                    "type": param_type,
                                    "items": {
                                        "type": item_type
                                    }
                                }
                        else:
                            # Handle non-array type parameters
                            properties[param_name] = {
                                "description": param_desc,
                                "type": param_type
                            }

                openai_function_schema = {
                    "name": f'{server.name}__{tool.name}',
                    "description": tool.description,
                    "parameters": {
                        "type": "object",
                        "properties": properties,
                        "required": required
                    }
                }
                openai_tools.append({
                    "type": "function",
                    "function": openai_function_schema,
                })
            logging.info(f"✅ server #{i + 1} ({server.name}) connected success，tools: {len(tools)}")

        except Exception as e:
            logging.error(f"❌ server #{i + 1} ({server.name}) connect fail: {e}")
            return []

    return openai_tools


async def mcp_tool_desc_transform(tools: List[str] = None,mcp_config: Dict[str, Any] = None) -> List[Dict[str, Any]]:
    """Default implement transform framework standard protocol to openai protocol of tool description."""
    config = {}
    global MCP_SERVERS_CONFIG
    def _replace_env_variables(config):
        if isinstance(config, dict):
            for key, value in config.items():
                if isinstance(value, str) and value.startswith("${") and value.endswith("}"):
                    env_var_name = value[2:-1]
                    config[key] = os.getenv(env_var_name, value)
                    logging.info(f"Replaced {value} with {config[key]}")
                elif isinstance(value, dict) or isinstance(value, list):
                    _replace_env_variables(value)
        elif isinstance(config, list):
            for index, item in enumerate(config):
                if isinstance(item, str) and item.startswith("${") and item.endswith("}"):
                    env_var_name = item[2:-1]
                    config[index] = os.getenv(env_var_name, item)
                    logging.info(f"Replaced {item} with {config[index]}")
                elif isinstance(item, dict) or isinstance(item, list):
                    _replace_env_variables(item)

    if mcp_config:
        try:
            config = mcp_config
            MCP_SERVERS_CONFIG = config
        except Exception as e:
            logging.error(f"mcp_config error: {e}")
            return []
    else:
        # Priority given to the running path.
        config_path = find_file(filename='mcp.json')
        if not os.path.exists(config_path):
            current_dir = os.path.dirname(os.path.abspath(__file__))
            config_path = os.path.normpath(os.path.join(current_dir, "../config/mcp.json"))
        logger.info(f"mcp conf path: {config_path}")

        if not os.path.exists(config_path):
            logging.info(f"mcp config is not exist: {config_path}")
            return []

        try:
            with open(config_path, "r") as f:
                config = json.load(f)
        except Exception as e:
            logging.info(f"load config fail: {e}")
            return []
        _replace_env_variables(config)

        MCP_SERVERS_CONFIG = config

    mcp_servers_config = config.get("mcpServers", {})

    server_configs = []
    for server_name, server_config in mcp_servers_config.items():
        # Skip disabled servers
        if server_config.get("disabled", False):
            continue

        if tools is None or server_name in tools:
            # Handle SSE server
            if "url" in server_config:
                server_configs.append({
                    "name": "mcp__" + server_name,
                    "type": "sse",
                    "params": {"url": server_config["url"]}
                })
            # Handle stdio server
            elif "command" in server_config:
                server_configs.append({
                    "name": "mcp__" + server_name,
                    "type": "stdio",
                    "params": {
                        "command": server_config["command"],
                        "args": server_config.get("args", []),
                        "env": server_config.get("env", {}),
                        "cwd": server_config.get("cwd"),
                        "encoding": server_config.get("encoding", "utf-8"),
                        "encoding_error_handler": server_config.get("encoding_error_handler", "strict")
                    }
                })

    if not server_configs:
        return []

    async with AsyncExitStack() as stack:
        servers = []
        for server_config in server_configs:
            try:
                if server_config["type"] == "sse":
                    server = MCPServerSse(
                        name=server_config["name"],
                        params=server_config["params"]
                    )
                elif server_config["type"] == "stdio":
                    from aworld.mcp_client.server import MCPServerStdio
                    server = MCPServerStdio(
                        name=server_config["name"],
                        params=server_config["params"]
                    )
                else:
                    logging.warning(f"Unsupported MCP server type: {server_config['type']}")
                    continue

                server = await stack.enter_async_context(server)
                servers.append(server)
            except BaseException as err:
                # single
                logging.error(
                    f"Failed to get tools for MCP server '{server_config['name']}'.\n"
                    f"Error: {err}\n"
                    f"Traceback:\n{traceback.format_exc()}"
                )

        openai_tools = await run(servers)

    return openai_tools



async def sandbox_mcp_tool_desc_transform(tools: List[str] = None,mcp_config: Dict[str, Any] = None) -> List[Dict[str, Any]]:
    # todo sandbox mcp_config get from registry

    if not mcp_config:
        return None
    config = mcp_config
    mcp_servers_config = config.get("mcpServers", {})
    server_configs = []
    openai_tools = []
    mcp_openai_tools = []

    for server_name, server_config in mcp_servers_config.items():
        # Skip disabled servers
        if server_config.get("disabled", False):
            continue

        if tools is None or server_name in tools:
            # Handle SSE server
            if "api" == server_config.get("type", ""):
                api_result = requests.get(server_config["url"]+"/list_tools")
                try:
                    if not api_result or not api_result.text:
                        continue
                        # return None
                    data = json.loads(api_result.text)
                    if not data or not data.get("tools"):
                        continue
                    for item in data.get("tools"):
                        tmp_function = {
                            "type": "function",
                            "function": {
                                "name": "mcp__" + server_name + "__" + item["name"],
                                "description": item["description"],
                                "parameters": {
                                    **item["parameters"],
                                    "properties": {k: v for k, v in item["parameters"].get("properties", {}).items()
                                                   if
                                                   'default' not in v}
                                }
                            }
                        }
                        openai_tools.append(tmp_function)
                except Exception as e:
                    logging.warning(f"server_name:{server_name} translate failed: {e}")
            elif "sse" == server_config.get("type", ""):
                server_configs.append({
                    "name": "mcp__" + server_name,
                    "type": "sse",
                    "params": {
                        "url": server_config["url"],
                        "headers": server_config.get("headers")
                    }
                })
            # Handle stdio server
            else:
            #elif "stdio" == server_config.get("type", ""):
                server_configs.append({
                    "name": "mcp__" + server_name,
                    "type": "stdio",
                    "params": {
                        "command": server_config["command"],
                        "args": server_config.get("args", []),
                        "env": server_config.get("env", {}),
                        "cwd": server_config.get("cwd"),
                        "encoding": server_config.get("encoding", "utf-8"),
                        "encoding_error_handler": server_config.get("encoding_error_handler", "strict")
                    }
                })

    if not server_configs:
        return openai_tools

    async with AsyncExitStack() as stack:
        servers = []
        for server_config in server_configs:
            try:
                if server_config["type"] == "sse":
                    server = MCPServerSse(
                        name=server_config["name"],
                        params=server_config["params"]
                    )
                elif server_config["type"] == "stdio":
                    server = MCPServerStdio(
                        name=server_config["name"],
                        params=server_config["params"]
                    )
                else:
                    logging.warning(f"Unsupported MCP server type: {server_config['type']}")
                    continue

                server = await stack.enter_async_context(server)
                servers.append(server)
            except BaseException as err:
                # single
                logging.error(
                    f"Failed to get tools for MCP server '{server_config['name']}'.\n"
                    f"Error: {err}\n"
                )

        mcp_openai_tools = await run(servers)

    if mcp_openai_tools:
        openai_tools.extend(mcp_openai_tools)

    return openai_tools


async def call_api(
        server_name: str,
        tool_name: str,
        parameter: Dict[str, Any] = None,
        mcp_config: Dict[str, Any] = None,
) -> ActionResult:
    """Specifically handle API type server calls

    Args:
        server_name: Server name
        tool_name: Tool name
        parameter: Parameters
        mcp_config: MCP configuration

    Returns:
        ActionResult: Call result
    """
    action_result = ActionResult(
        content="",
        keep=True
    )

    if not mcp_config or mcp_config.get("mcpServers") is None:
        return action_result

    mcp_servers = mcp_config.get("mcpServers")
    if not mcp_servers.get(server_name):
        return action_result

    server_config = mcp_servers.get(server_name)
    if "api" != server_config.get("type", ""):
        logging.warning(f"Server {server_name} is not API type, should use call_tool instead")
        return action_result

    try:
        headers = {
            "Content-Type": "application/json"
        }
        response = requests.post(
            url=server_config["url"] + "/" + tool_name,
            headers=headers,
            json=parameter
        )
        action_result = ActionResult(
            content=response.text,
            keep=True
        )
    except Exception as e:
        logging.warning(f"call_api ({server_name})({tool_name}) failed: {e}")
        action_result = ActionResult(
            content=f"Error calling API: {str(e)}",
            keep=True
        )

    return action_result



async def get_server_instance(server_name: str, mcp_config: Dict[str, Any] = None) -> Any:
    """Get server instance, create a new one if it doesn't exist

    Args:
        server_name: Server name
        mcp_config: MCP configuration

    Returns:
        Server instance or None (if creation fails)
    """
    if not mcp_config or mcp_config.get("mcpServers") is None:
        return None

    mcp_servers = mcp_config.get("mcpServers")
    if not mcp_servers.get(server_name):
        return None

    server_config = mcp_servers.get(server_name)
    try:
        # API type servers use special handling, no need for persistent connections
        # Note: We've already handled API type in McpServers.call_tool method
        # Here we don't return None, but let the caller handle it
        if "api" == server_config.get("type", ""):
            logging.info(f"API server {server_name} doesn't need persistent connection")
            return None
        elif "sse" == server_config.get("type", ""):
            server = MCPServerSse(
                name=server_name,
                params={
                    "url": server_config["url"],
                    "headers": server_config.get("headers"),
                    "timeout": server_config.get("timeout", 5.0),
                    "sse_read_timeout": server_config.get("sse_read_timeout", 300.0)
                }
            )
            await server.connect()
            logging.info(f"Successfully connected to SSE server: {server_name}")
            return server
        else:  # stdio type
            params = {
                "command": server_config["command"],
                "args": server_config.get("args", []),
                "env": server_config.get("env", {}),
                "cwd": server_config.get("cwd"),
                "encoding": server_config.get("encoding", "utf-8"),
                "encoding_error_handler": server_config.get("encoding_error_handler", "strict")
            }
            server = MCPServerStdio(
                name=server_name,
                params=params
            )
            await server.connect()
            logging.info(f"Successfully connected to stdio server: {server_name}")
            return server
    except Exception as e:
        logging.warning(f"Failed to create server instance for {server_name}: {e}")
        return None


async def cleanup_server(server):
    """Clean up server connection

    Args:
        server: Server instance
    """
    try:
        if hasattr(server, 'cleanup'):
            await server.cleanup()
        elif hasattr(server, 'close'):
            await server.close()
        logging.info(f"Successfully cleaned up server: {getattr(server, 'name', 'unknown')}")
    except Exception as e:
        logging.warning(f"Failed to cleanup server: {e}")



import abc
import asyncio
import logging
import uuid
from typing import Dict, List, Any, Optional

from aworld.sandbox.api.setup import SandboxSetup
from aworld.sandbox.models import SandboxStatus, SandboxEnvType, SandboxInfo
from aworld.sandbox.run.mcp_servers import McpServers


class Sandbox(SandboxSetup):
    """
    Sandbox abstract base class that defines the interface for all sandbox implementations.
    A sandbox provides an isolated environment for executing code and operations.
    """
    
    default_sandbox_timeout = 3000
    
    @property
    def sandbox_id(self) -> str:
        """
        Returns the unique identifier of the sandbox.
        """
        return self._sandbox_id

    @property
    def status(self) -> SandboxStatus:
        """
        Returns the current status of the sandbox.
        """
        return self._status

    @property
    def timeout(self) -> int:
        """
        Returns the timeout value for sandbox operations.
        """
        return self._timeout

    @property
    def metadata(self) -> Dict[str, str]:
        """
        Returns the sandbox metadata.
        """
        return self._metadata

    @property
    def env_type(self) -> SandboxEnvType:
        """
        Returns the environment type of the sandbox.
        """
        return self._env_type

    @property
    def mcp_config(self) -> Any:
        """
        Returns the MCP configuration.
        """
        return self._mcp_config

    @property
    def mcp_servers(self) -> List[str]:
        """
        Returns the list of MCP servers.
        """
        return self._mcp_servers
    
    @property
    @abc.abstractmethod
    def mcpservers(self) -> McpServers:
        """
        Module for running MCP in the sandbox.
        
        Returns:
            McpServers: The MCP servers instance.
        """
        pass

    def __init__(
            self,
            sandbox_id: Optional[str] = None,
            env_type: Optional[int] = None,
            metadata: Optional[Dict[str, str]] = None,
            timeout: Optional[int] = None,
            mcp_servers: Optional[List[str]] = None,
            mcp_config: Optional[Any] = None,
    ):
        """
        Initialize a new Sandbox instance.
        
        Args:
            sandbox_id: Unique identifier for the sandbox. If None, one will be generated.
            env_type: The environment type (LOCAL, K8S, SUPERCOMPUTER).
            metadata: Additional metadata for the sandbox.
            timeout: Timeout for sandbox operations.
            mcp_servers: List of MCP servers to use.
            mcp_config: Configuration for MCP servers.
        """
        # Initialize basic attributes
        self._sandbox_id = sandbox_id or str(uuid.uuid4())
        self._status = SandboxStatus.INIT
        self._timeout = timeout or self.default_sandbox_timeout
        self._metadata = metadata or {}
        self._env_type = env_type or SandboxEnvType.LOCAL
        self._mcp_servers = mcp_servers or []
        self._mcp_config = mcp_config or {}
        
    @abc.abstractmethod
    def get_info(self) -> SandboxInfo:
        """
        Returns information about the sandbox.
        
        Returns:
            SandboxInfo: Information about the sandbox.
        """
        pass
    
    @abc.abstractmethod
    async def remove(self) -> bool:
        """
        Remove the sandbox and clean up all resources.
        
        Returns:
            bool: True if removal was successful, False otherwise.
        """
        pass
    
    @abc.abstractmethod
    async def cleanup(self) -> bool:
        """
        Clean up the sandbox resources.
        
        Returns:
            bool: True if cleanup was successful, False otherwise.
        """
        pass
    
    def __del__(self):
        """
        Ensure resources are cleaned up when the object is garbage collected.
        """
        try:
            # Handle the case where an event loop already exists
            try:
                loop = asyncio.get_running_loop()
                logging.warning("Cannot clean up sandbox in __del__ when event loop is already running")
                return
            except RuntimeError:
                # No running event loop, create a new one
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
                loop.run_until_complete(self.cleanup())
                loop.close()
        except Exception as e:
            logging.warning(f"Failed to cleanup sandbox resources during garbage collection: {e}")


import os
import json
import time
import threading
from datetime import datetime
from abc import ABC, abstractmethod
from collections import defaultdict
from pydantic import BaseModel
from typing import Optional, Dict, Any, Union
from opentelemetry.sdk.trace import Span, SpanContext
from opentelemetry.sdk.trace.export import SpanExporter
from aworld.logs.util import logger
from aworld.trace.constants import ATTRIBUTES_MESSAGE_RUN_TYPE_KEY, RunType


class SpanStatus(BaseModel):
    code: str = "UNSET"
    description: Optional[str] = None


class SpanModel(BaseModel):
    trace_id: str
    span_id: str
    name: str
    start_time: str
    end_time: str
    duration_ms: float
    attributes: Dict[str, Any]
    status: SpanStatus
    parent_id: Optional[str]
    children: list['SpanModel'] = []
    run_type: Optional[str] = RunType.OTHER.value

    @staticmethod
    def from_span(span):
        start_timestamp = span.start_time / 1e9
        end_timestamp = span.end_time / 1e9
        start_ms = int((span.start_time % 1e9) / 1e6)
        end_ms = int((span.end_time % 1e9) / 1e6)

        return SpanModel(
            trace_id=f"{span.get_span_context().trace_id:032x}",
            span_id=SpanModel.get_span_id(span),
            name=span.name,
            start_time=time.strftime(
                '%Y-%m-%d %H:%M:%S', time.localtime(start_timestamp)) + f'.{start_ms:03d}',
            end_time=time.strftime(
                '%Y-%m-%d %H:%M:%S', time.localtime(end_timestamp)) + f'.{end_ms:03d}',
            duration_ms=(span.end_time - span.start_time)/1e6,
            attributes={k: v for k, v in span.attributes.items()},
            status=SpanStatus(
                code=str(
                    span.status.status_code) if span.status.status_code else "UNSET",
                description=span.status.description or None
            ),
            parent_id=SpanModel.get_span_id(
                span.parent) if span.parent else None,
            run_type=span.attributes.get(
                ATTRIBUTES_MESSAGE_RUN_TYPE_KEY, RunType.OTHER.value)
        )

    @staticmethod
    def get_span_id(span: Union[Span, SpanContext]):
        if isinstance(span, SpanContext):
            return f"{span.span_id:016x}"
        return f"{span.get_span_context().span_id:016x}"


class TraceStorage(ABC):
    """
    Storage for traces.
    """
    @abstractmethod
    def add_span(self, span: Span) -> None:
        """
        Add a span to the storage.
        """

    @abstractmethod
    def get_all_traces(self) -> list[str]:
        """
        Get all trace ids.
        """

    @abstractmethod
    def get_all_spans(self, trace_id) -> list[SpanModel]:
        """
        Get all spans of a trace.
        """


class InMemoryStorage(TraceStorage):
    """
    In-memory storage for spans.
    """

    def __init__(self,  max_traces=1000):
        self._traces = defaultdict(list)
        self._trace_order = []
        self.max_traces = max_traces

    def add_span(self, span: Span):
        trace_id = f"{span.get_span_context().trace_id:032x}"
        if trace_id not in self._traces:
            self._trace_order.append(trace_id)
            if len(self._trace_order) > self.max_traces:
                oldest_trace = self._trace_order.pop(0)
                del self._traces[oldest_trace]
        self._traces[trace_id].append(SpanModel.from_span(span))

    def get_all_traces(self):
        return list(self._traces.keys())

    def get_all_spans(self, trace_id):
        return self._traces.get(trace_id, [])


class InMemoryWithPersistStorage(TraceStorage):
    """
    In-memory storage for spans with optimized disk persistence.
    """

    def __init__(self, storage_dir: str = "./trace_data"):
        self._traces = defaultdict(list)
        self._pending_spans = []
        self.storage_dir = os.path.abspath(storage_dir)
        os.makedirs(self.storage_dir, exist_ok=True)
        self._lock = threading.Lock()
        self._persist_thread = None
        self._load_today_traces()
        self.current_filename = None

    def _get_today_filename(self):
        if not self.current_filename:
            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
            self.current_filename = f"trace_{timestamp}.json"
        return self.current_filename

    def _load_today_traces(self):
        today = datetime.now().strftime("%Y%m%d")
        for filename in os.listdir(self.storage_dir):
            if filename.startswith(f"trace_{today}") and filename.endswith(".json"):
                filepath = os.path.join(self.storage_dir, filename)
                try:
                    with self._lock, open(filepath, 'r') as f:
                        data = json.load(f)
                        for span_data in data:
                            trace_id = span_data.get("trace_id")
                            span_json = span_data.get("span")
                            self._traces[trace_id].append(
                                SpanModel.parse_raw(span_json))
                except Exception as e:
                    logger.error(
                        f"Error loading trace file {filename}: {str(e)}")

    def _start_persist_thread(self):
        if self._persist_thread is None:
            self._persist_thread = threading.Thread(
                target=self._persist_worker, daemon=True)
            self._persist_thread.start()

    def _persist_worker(self):
        while True:
            time.sleep(5)
            self._persist()

    def _persist(self):
        if not self._pending_spans:
            return

        temp_filepath = os.path.join(
            self.storage_dir, f"temp_{time.time_ns()}.json")
        final_filepath = os.path.join(
            self.storage_dir, self._get_today_filename())

        try:
            spans_to_persist = []
            with self._lock:
                spans_to_persist = self._pending_spans.copy()
                self._pending_spans.clear()

            if spans_to_persist:
                existing_data = []
                if os.path.exists(final_filepath):
                    try:
                        with open(final_filepath, 'r') as f:
                            existing_data = json.load(f)
                    except Exception as e:
                        logger.error(
                            f"Error reading existing trace file: {str(e)}")

                merged_spans = existing_data + spans_to_persist

                with open(temp_filepath, 'w') as f:
                    json.dump(merged_spans, f, default=str)
                os.replace(temp_filepath, final_filepath)
        except Exception as e:
            logger.error(f"Error persisting traces: {str(e)}")
            try:
                os.unlink(temp_filepath)
            except:
                pass

    def add_span(self, span: Span):
        span_model = SpanModel.from_span(span)
        with self._lock:
            self._traces[span_model.trace_id].append(span_model)
            self._pending_spans.append({
                "trace_id": span_model.trace_id,
                "span": span_model.json()
            })
        self._start_persist_thread()

    def get_all_traces(self):
        with self._lock:
            return list(self._traces.keys())

    def get_all_spans(self, trace_id):
        with self._lock:
            return self._traces.get(trace_id, [])


class InMemorySpanExporter(SpanExporter):
    """
    Span exporter that stores spans in memory.
    """

    def __init__(self, storage: TraceStorage):
        self._storage = storage

    def export(self, spans):
        for span in spans:
            self._storage.add_span(span)

    def shutdown(self):
        pass



# coding: utf-8
# Copyright (c) 2025 inclusionAI.
import importlib
import inspect
import os
import sys

from typing import Callable, Any, get_type_hints, get_origin, get_args

from pydantic import create_model, Field, BaseModel
from pydantic.fields import FieldInfo

from aworld.core.common import ToolActionInfo, ParamInfo
from aworld.core.tool.action import TOOL_ACTION
from aworld.core.tool.action_factory import ActionFactory
from aworld.core.tool.action_template import ACTION_TEMPLATE
from aworld.core.tool.base import ToolFactory
from aworld.core.tool.tool_template import TOOL_TEMPLATE
from aworld.logs.util import logger


def be_tool(
        tool_name: str = None,
        tool_desc: str = None,
        name: str = None,
        desc: str = None, **kwargs
) -> Callable[..., Any]:
    """Decorate a function to be a tool, auto register the tool and action with the parameters with the factory.

    Example:
        >>> @be_tool()
        >>> def example():
        >>>     return "example"

        # write name and description
        >>> @be_tool(name="param", desc="example param func")
        >>> def example_param(param: str):
        >>>     return param

        >>> @be_tool(tool_name='field_param_tool')
        >>> def example_param_field(param: str = Field(..., description="param")):
        >>>     return param

        >>> @be_tool(tool_name='field_param_tool')
        >>> def example_param_2(param: str = Field(..., description="param")):
        >>>     return param

    Args:
        tool_name: Optional name for the tool.
        tool_desc: Optional description for the tool.
        name: Optional name for the function.
        desc: Optional description of function.
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        # converts python function into a tool with associated actions
        function_to_tool(
            func,
            tool_name=tool_name,
            tool_desc=tool_desc,
            name=name,
            desc=desc,
            **kwargs
        )
        return func

    return decorator


def function_to_tool(
        func: Callable[..., Any],
        tool_name: str = None,
        tool_desc: str = None,
        name: str = None,
        desc: str = None,
        **kwargs
) -> None:
    """Transform a python function into a tool, and register to the factory.

    Generates necessary code files dynamically and manages tool and action registration.

    Args：
        func: An executable function.
        tool_name: The name of the tool that is transformed by the function.
        tool_desc: The description of the tool what it can do, default the same as tool_name.
        name: Alias name of function.
        desc: The description of the function what it can do, default the same as name.
    """
    tool_name = tool_name or name or func.__name__
    action_name = name or func.__name__

    if tool_name == "<lambda>" or action_name == "<lambda>":
        raise ValueError("You must provide a name for lambda functions")

    # async func, will use AsyncTool
    is_async = inspect.iscoroutinefunction(func)

    name = action_name
    if not inspect.iscoroutinefunction(func):
        name = f"async_func({action_name})"

    # build action
    if action_name not in ActionFactory:
        func_import = func.__module__
        if func_import == '__main__':
            path = inspect.getsourcefile(func)
            package = path.replace(os.getcwd(), '').replace('.py', '')
            if package[0] == '/':
                package = package[1:]
            func_import = f"from {package} "
        else:
            func_import = f"from {func_import} "
        con = ACTION_TEMPLATE.format(name=action_name,
                                     desc=desc if desc else action_name,
                                     tool_name=tool_name,
                                     func_import=func_import,
                                     func=func.__name__,
                                     call_func=name)
        with open(f"{action_name}.py", 'w+') as write:
            write.writelines(con)
        module = importlib.import_module(action_name)
        getattr(module, action_name)

        if not kwargs.get('keep_file', False):
            os.remove(f"{action_name}.py")
    else:
        logger.warning(f"{action_name} already register to the tool.")
        raise ValueError(f"{action_name} already register to a tool.")

    # build params info
    parameters = func_params(func)

    module_name = f'{tool_name}_action'
    if module_name not in sys.modules:
        # ToolAction process
        with open(f"{module_name}.py", 'w+') as write:
            write.writelines(TOOL_ACTION.format(name=tool_name))
        module = importlib.import_module(module_name)
        if not kwargs.get('keep_file', False):
            os.remove(f"{module_name}.py")
        tool_action_cls = getattr(module, f"{tool_name}Action")
    else:
        logger.info(f"{module_name} already exists in modules, reuse the tool action.")
        tool_action_cls = getattr(sys.modules[module_name], f"{tool_name}Action")

    params = {}
    if parameters:
        for k, v in parameters['properties'].items():
            params[k] = ParamInfo(name=k,
                                  type=v.get('type', 'string'),
                                  required=False if v.get('default') else True,
                                  default_value=v.get('default'),
                                  desc=v.get('description', k))

    setattr(tool_action_cls,
            action_name.upper(),
            ToolActionInfo(
                name=action_name,
                desc=desc if desc else action_name,
                input_params=params
            ))

    # build tool
    if tool_name not in ToolFactory:
        con = TOOL_TEMPLATE.format(name=tool_name,
                                   desc=tool_desc if tool_desc else tool_name,
                                   action=f"{tool_name}Action",
                                   action_import=f"from {tool_action_cls.__module__} import {tool_name}Action",
                                   cls='AsyncTool' if is_async else 'Tool',
                                   async_flag='async ' if is_async else '',
                                   async_underline='async_' if is_async else '',
                                   await_flag='await ' if is_async else '')

        if tool_name == action_name:
            tool_name += '_tool'

        with open(f"{tool_name}.py", 'w+') as write:
            write.writelines(con)
        importlib.import_module(tool_name)

        if not kwargs.get('keep_file', False):
            os.remove(f"{tool_name}.py")


def func_params(func: Callable[..., Any]):
    """Extracts parameter information from the function.

    Args:
        func: An executable function.

    Returns:
        JSON schema of the function input parameters.
    """
    sig = inspect.signature(func)
    type_hints = get_type_hints(func)
    filtered_params = []

    # The function must have a return value
    if sig.return_annotation == inspect.Parameter.empty:
        raise RuntimeError(f"{func} no return value, preferably a string.")

    for name, param in sig.parameters.items():
        filtered_params.append((name, param))

    fields: dict[str, Any] = {}
    param_descs = {}

    for name, param in filtered_params:
        ann = type_hints.get(name, param.annotation)
        default = param.default
        def_desc = None
        if hasattr(default, 'description'):
            def_desc = default.description
        field_description = param_descs.get(name, def_desc)
        
        if isinstance(default, FieldInfo):
            default = default.default

        # If there's no type hint, assume `Any`
        if ann == inspect.Parameter.empty:
            ann = Any

        # Handle different parameter kinds
        if param.kind == param.VAR_POSITIONAL:
            # e.g. *args: extend positional args
            if get_origin(ann) is tuple:
                # e.g. def foo(*args: tuple[int, ...]) -> treat as List[int]
                args_of_tuple = get_args(ann)
                if len(args_of_tuple) == 2 and args_of_tuple[1] is Ellipsis:
                    ann = list[args_of_tuple[0]]  # type: ignore
                else:
                    ann = list[Any]
            else:
                # If user wrote *args: int, treat as List[int]
                ann = list[ann]  # type: ignore

            # Default factory to empty list
            fields[name] = (
                ann,
                Field(default_factory=list, description=field_description),  # type: ignore
            )
        elif param.kind == param.VAR_KEYWORD:
            # **kwargs handling
            if get_origin(ann) is dict:
                # e.g. def foo(**kwargs: dict[str, int])
                dict_args = get_args(ann)
                if len(dict_args) == 2:
                    ann = dict[dict_args[0], dict_args[1]]  # type: ignore
                else:
                    ann = dict[str, Any]
            else:
                # e.g. def foo(**kwargs: int) -> Dict[str, int]
                ann = dict[str, ann]

            fields[name] = (
                ann,
                Field(default_factory=dict, description=field_description),  # type: ignore
            )
        else:
            if default == inspect.Parameter.empty:
                # Required field
                fields[name] = (ann, Field(..., description=field_description))
            else:
                # Parameter with a default value
                fields[name] = (ann, Field(default=default, description=field_description))

    dynamic_model = create_model(f"{func.__name__}".upper(), __base__=BaseModel, **fields)
    json_schema = dynamic_model.model_json_schema()
    logger.debug(f"{func} parameters schema: {json_schema}")
    return json_schema



import os
import click
import streamlit.web.bootstrap as bootstrap


@click.group()
def main(*args, **kwargs):
    print(
        """AWorld CLI Help:
    aworld run: run aworld web agent
    aworld help: show help"""
    )


@main.command("run")
@click.argument("args", nargs=-1)
def main_run(args=None, **kwargs):
    script = os.path.join(os.path.dirname(os.path.abspath(__file__)), "main.py")
    bootstrap.run(script, False, args[1:], flag_options=kwargs)


if __name__ == "__main__":
    main()



import os
from typing import Any, Dict, List, Generator, AsyncGenerator

from openai import OpenAI, AsyncOpenAI

from aworld.config.conf import ClientType
from aworld.core.llm_provider_base import LLMProviderBase
from aworld.models.llm_http_handler import LLMHTTPHandler
from aworld.models.model_response import ModelResponse, LLMResponseError
from aworld.logs.util import logger
from aworld.models.utils import usage_process


class OpenAIProvider(LLMProviderBase):
    """OpenAI provider implementation.
    """

    def _init_provider(self):
        """Initialize OpenAI provider.
        
        Returns:
            OpenAI provider instance.
        """
        # Get API key
        api_key = self.api_key
        if not api_key:
            env_var = "OPENAI_API_KEY"
            api_key = os.getenv(env_var, "")
            if not api_key:
                raise ValueError(
                    f"OpenAI API key not found, please set {env_var} environment variable or provide it in the parameters")
        base_url = self.base_url
        if not base_url:
            base_url = os.getenv("OPENAI_ENDPOINT", "https://api.openai.com/v1")

        self.is_http_provider = False
        if self.kwargs.get("client_type", ClientType.SDK) == ClientType.HTTP:
            logger.info(f"Using HTTP provider for OpenAI")
            self.http_provider = LLMHTTPHandler(
                base_url=base_url,
                api_key=api_key,
                model_name=self.model_name,
                max_retries=self.kwargs.get("max_retries", 3)
            )
            self.is_http_provider = True
            return self.http_provider
        else:
            return OpenAI(
                api_key=api_key,
                base_url=base_url,
                timeout=self.kwargs.get("timeout", 180),
                max_retries=self.kwargs.get("max_retries", 3)
            )

    def _init_async_provider(self):
        """Initialize async OpenAI provider.

        Returns:
            Async OpenAI provider instance.
        """
        # Get API key
        api_key = self.api_key
        if not api_key:
            env_var = "OPENAI_API_KEY"
            api_key = os.getenv(env_var, "")
            if not api_key:
                raise ValueError(
                    f"OpenAI API key not found, please set {env_var} environment variable or provide it in the parameters")
        base_url = self.base_url
        if not base_url:
            base_url = os.getenv("OPENAI_ENDPOINT", "https://api.openai.com/v1")

        return AsyncOpenAI(
            api_key=api_key,
            base_url=base_url,
            timeout=self.kwargs.get("timeout", 180),
            max_retries=self.kwargs.get("max_retries", 3)
        )

    @classmethod
    def supported_models(cls) -> list[str]:
        return ["gpt-4o", "gpt-4", "gpt-3.5-turbo", "o3-mini", "gpt-4o-mini", "deepseek-chat", "deepseek-reasoner",
                r"qwq-.*", r"qwen-.*"]

    def preprocess_messages(self, messages: List[Dict[str, str]]) -> List[Dict[str, str]]:
        """Preprocess messages, use OpenAI format directly.

        Args:
            messages: OpenAI format message list.

        Returns:
            Processed message list.
        """
        for message in messages:
            if message["role"] == "assistant" and "tool_calls" in message and message["tool_calls"]:
                if message["content"] is None: message["content"] = ""
                for tool_call in message["tool_calls"]:
                    if "function" not in tool_call and "name" in tool_call and "arguments" in tool_call:
                        tool_call["function"] = {"name": tool_call["name"], "arguments": tool_call["arguments"]}

        return messages

    def postprocess_response(self, response: Any) -> ModelResponse:
        """Process OpenAI response.

        Args:
            response: OpenAI response object.

        Returns:
            ModelResponse object.
            
        Raises:
            LLMResponseError: When LLM response error occurs.
        """
        if ((not isinstance(response, dict) and (not hasattr(response, 'choices') or not response.choices))
                or (isinstance(response, dict) and not response.get("choices"))):
            error_msg = ""
            if hasattr(response, 'error') and response.error and isinstance(response.error, dict):
                error_msg = response.error.get('message', '')
            elif hasattr(response, 'msg'):
                error_msg = response.msg

            raise LLMResponseError(
                error_msg if error_msg else "Unknown error",
                self.model_name or "unknown",
                response
            )

        return ModelResponse.from_openai_response(response)

    def postprocess_stream_response(self, chunk: Any) -> ModelResponse:
        """Process OpenAI streaming response chunk.

        Args:
            chunk: OpenAI response chunk.

        Returns:
            ModelResponse object.
            
        Raises:
            LLMResponseError: When LLM response error occurs.
        """
        # Check if chunk contains error
        if hasattr(chunk, 'error') or (isinstance(chunk, dict) and chunk.get('error')):
            error_msg = chunk.error if hasattr(chunk, 'error') else chunk.get('error', 'Unknown error')
            raise LLMResponseError(
                error_msg,
                self.model_name or "unknown",
                chunk
            )

        # process tool calls
        if (hasattr(chunk, 'choices') and chunk.choices and chunk.choices[0].delta and chunk.choices[0].delta.tool_calls) or (
                isinstance(chunk, dict) and chunk.get("choices") and chunk["choices"] and chunk["choices"][0].get("delta", {}).get("tool_calls")):
            tool_calls = chunk.choices[0].delta.tool_calls if hasattr(chunk, 'choices') else chunk["choices"][0].get("delta", {}).get("tool_calls")

            for tool_call in tool_calls:
                index = tool_call.index if hasattr(tool_call, 'index') else tool_call["index"]
                func_name = tool_call.function.name if hasattr(tool_call, 'function') else tool_call.get("function", {}).get("name")
                func_args = tool_call.function.arguments if hasattr(tool_call, 'function') else tool_call.get("function", {}).get("arguments")
                if index >= len(self.stream_tool_buffer):
                    self.stream_tool_buffer.append({
                        "id": tool_call.id if hasattr(tool_call, 'id') else tool_call.get("id"),
                        "type": "function",
                        "function": {
                            "name": func_name,
                            "arguments": func_args
                        }
                    })
                else:
                    self.stream_tool_buffer[index]["function"]["arguments"] += func_args
            processed_chunk = chunk
            if hasattr(processed_chunk, 'choices'):
                processed_chunk.choices[0].delta.tool_calls = None
            else:
                processed_chunk["choices"][0]["delta"]["tool_calls"] = None
            resp = ModelResponse.from_openai_stream_chunk(processed_chunk)
            if (not resp.content and not resp.usage.get("total_tokens", 0)):
                return None
        if (hasattr(chunk, 'choices') and chunk.choices and chunk.choices[0].finish_reason) or (
                isinstance(chunk, dict) and chunk.get("choices") and chunk["choices"] and chunk["choices"][0].get(
            "finish_reason")):
            finish_reason = chunk.choices[0].finish_reason if hasattr(chunk, 'choices') else chunk["choices"][0].get(
                "finish_reason")
            if self.stream_tool_buffer:
                tool_call_chunk = {
                    "id": chunk.id if hasattr(chunk, 'id') else chunk.get("id"),
                    "model": chunk.model if hasattr(chunk, 'model') else chunk.get("model"),
                    "object": chunk.object if hasattr(chunk, 'object') else chunk.get("object"),
                    "choices": [
                        {
                            "delta": {
                                "role": "assistant",
                                "content": "",
                                "tool_calls": self.stream_tool_buffer
                            }
                        }
                    ]
                }
                self.stream_tool_buffer = []
                return ModelResponse.from_openai_stream_chunk(tool_call_chunk)

        return ModelResponse.from_openai_stream_chunk(chunk)

    def completion(self,
                   messages: List[Dict[str, str]],
                   temperature: float = 0.0,
                   max_tokens: int = None,
                   stop: List[str] = None,
                   **kwargs) -> ModelResponse:
        """Synchronously call OpenAI to generate response.
        
        Args:
            messages: Message list.
            temperature: Temperature parameter.
            max_tokens: Maximum number of tokens to generate.
            stop: List of stop sequences.
            **kwargs: Other parameters.

        Returns:
            ModelResponse object.
            
        Raises:
            LLMResponseError: When LLM response error occurs.
        """
        if not self.provider:
            raise RuntimeError(
                "Sync provider not initialized. Make sure 'sync_enabled' parameter is set to True in initialization.")

        processed_messages = self.preprocess_messages(messages)

        try:
            openai_params = self.get_openai_params(processed_messages, temperature, max_tokens, stop, **kwargs)
            if self.is_http_provider:
                response = self.http_provider.sync_call(openai_params)
            else:
                response = self.provider.chat.completions.create(**openai_params)

            if (hasattr(response, 'code') and response.code != 0) or (
                    isinstance(response, dict) and response.get("code", 0) != 0):
                error_msg = getattr(response, 'msg', 'Unknown error')
                logger.warn(f"API Error: {error_msg}")
                raise LLMResponseError(error_msg, kwargs.get("model_name", self.model_name or "unknown"), response)

            if not response:
                raise LLMResponseError("Empty response", kwargs.get("model_name", self.model_name or "unknown"))

            resp = self.postprocess_response(response)
            usage_process(resp.usage)
            return resp
        except Exception as e:
            if isinstance(e, LLMResponseError):
                raise e
            logger.warn(f"Error in OpenAI completion: {e}")
            raise LLMResponseError(str(e), kwargs.get("model_name", self.model_name or "unknown"))

    def stream_completion(self,
                          messages: List[Dict[str, str]],
                          temperature: float = 0.0,
                          max_tokens: int = None,
                          stop: List[str] = None,
                          **kwargs) -> Generator[ModelResponse, None, None]:
        """Synchronously call OpenAI to generate streaming response.
        
        Args:
            messages: Message list.
            temperature: Temperature parameter.
            max_tokens: Maximum number of tokens to generate.
            stop: List of stop sequences.
            **kwargs: Other parameters.

        Returns:
            Generator yielding ModelResponse chunks.
            
        Raises:
            LLMResponseError: When LLM response error occurs.
        """
        if not self.provider:
            raise RuntimeError(
                "Sync provider not initialized. Make sure 'sync_enabled' parameter is set to True in initialization.")

        processed_messages = self.preprocess_messages(messages)
        usage={
            "completion_tokens": 0,
            "prompt_tokens": 0,
            "total_tokens": 0
        }

        try:
            openai_params = self.get_openai_params(processed_messages, temperature, max_tokens, stop, **kwargs)
            openai_params["stream"] = True
            if self.is_http_provider:
                response_stream = self.http_provider.sync_stream_call(openai_params)
            else:
                response_stream = self.provider.chat.completions.create(**openai_params)

            for chunk in response_stream:
                if not chunk:
                    continue
                resp = self.postprocess_stream_response(chunk)
                if resp:
                    self._accumulate_chunk_usage(usage, resp.usage)
                    yield resp
            usage_process(usage)

        except Exception as e:
            logger.warn(f"Error in stream_completion: {e}")
            raise LLMResponseError(str(e), kwargs.get("model_name", self.model_name or "unknown"))

    async def astream_completion(self,
                                 messages: List[Dict[str, str]],
                                 temperature: float = 0.0,
                                 max_tokens: int = None,
                                 stop: List[str] = None,
                                 **kwargs) -> AsyncGenerator[ModelResponse, None]:
        """Asynchronously call OpenAI to generate streaming response.
        
        Args:
            messages: Message list.
            temperature: Temperature parameter.
            max_tokens: Maximum number of tokens to generate.
            stop: List of stop sequences.
            **kwargs: Other parameters.

        Returns:
            AsyncGenerator yielding ModelResponse chunks.
            
        Raises:
            LLMResponseError: When LLM response error occurs.
        """
        if not self.async_provider:
            raise RuntimeError(
                "Async provider not initialized. Make sure 'async_enabled' parameter is set to True in initialization.")

        processed_messages = self.preprocess_messages(messages)
        usage = {
            "completion_tokens": 0,
            "prompt_tokens": 0,
            "total_tokens": 0
        }

        try:
            openai_params = self.get_openai_params(processed_messages, temperature, max_tokens, stop, **kwargs)
            openai_params["stream"] = True

            if self.is_http_provider:
                async for chunk in self.http_provider.async_stream_call(openai_params):
                    if not chunk:
                        continue
                    resp = self.postprocess_stream_response(chunk)
                    self._accumulate_chunk_usage(usage, resp.usage)
                    yield resp
            else:
                response_stream = await self.async_provider.chat.completions.create(**openai_params)
                async for chunk in response_stream:
                    if not chunk:
                        continue
                    resp = self.postprocess_stream_response(chunk)
                    if resp:
                        self._accumulate_chunk_usage(usage, resp.usage)
                        yield resp
            usage_process(usage)

        except Exception as e:
            logger.warn(f"Error in astream_completion: {e}")
            raise LLMResponseError(str(e), kwargs.get("model_name", self.model_name or "unknown"))

    async def acompletion(self,
                          messages: List[Dict[str, str]],
                          temperature: float = 0.0,
                          max_tokens: int = None,
                          stop: List[str] = None,
                          **kwargs) -> ModelResponse:
        """Asynchronously call OpenAI to generate response.
        
        Args:
            messages: Message list.
            temperature: Temperature parameter.
            max_tokens: Maximum number of tokens to generate.
            stop: List of stop sequences.
            **kwargs: Other parameters.

        Returns:
            ModelResponse object.
            
        Raises:
            LLMResponseError: When LLM response error occurs.
        """
        if not self.async_provider:
            raise RuntimeError(
                "Async provider not initialized. Make sure 'async_enabled' parameter is set to True in initialization.")

        processed_messages = self.preprocess_messages(messages)

        try:
            openai_params = self.get_openai_params(processed_messages, temperature, max_tokens, stop, **kwargs)
            if self.is_http_provider:
                response = await self.http_provider.async_call(openai_params)
            else:
                response = await self.async_provider.chat.completions.create(**openai_params)

            if (hasattr(response, 'code') and response.code != 0) or (
                    isinstance(response, dict) and response.get("code", 0) != 0):
                error_msg = getattr(response, 'msg', 'Unknown error')
                logger.warn(f"API Error: {error_msg}")
                raise LLMResponseError(error_msg, kwargs.get("model_name", self.model_name or "unknown"), response)

            if not response:
                raise LLMResponseError("Empty response", kwargs.get("model_name", self.model_name or "unknown"))

            resp = self.postprocess_response(response)
            usage_process(resp.usage)
            return resp
        except Exception as e:
            if isinstance(e, LLMResponseError):
                raise e
            logger.warn(f"Error in acompletion: {e}")
            raise LLMResponseError(str(e), kwargs.get("model_name", self.model_name or "unknown"))

    def get_openai_params(self,
                          messages: List[Dict[str, str]],
                          temperature: float = 0.0,
                          max_tokens: int = None,
                          stop: List[str] = None,
                          **kwargs) -> Dict[str, Any]:
        openai_params = {
            "model": kwargs.get("model_name", self.model_name or ""),
            "messages": messages,
            "temperature": temperature,
            "max_tokens": max_tokens,
            "stop": stop
        }

        supported_params = [
            "frequency_penalty", "logit_bias", "logprobs", "top_logprobs",
            "presence_penalty", "response_format", "seed", "stream", "top_p",
            "user", "function_call", "functions", "tools", "tool_choice"
        ]

        for param in supported_params:
            if param in kwargs:
                openai_params[param] = kwargs[param]

        return openai_params


class AzureOpenAIProvider(OpenAIProvider):
    """Azure OpenAI provider implementation.
    """

    def _init_provider(self):
        """Initialize Azure OpenAI provider.

        Returns:
            Azure OpenAI provider instance.
        """
        from langchain_openai import AzureChatOpenAI

        # Get API key
        api_key = self.api_key
        if not api_key:
            env_var = "AZURE_OPENAI_API_KEY"
            api_key = os.getenv(env_var, "")
            if not api_key:
                raise ValueError(
                    f"Azure OpenAI API key not found, please set {env_var} environment variable or provide it in the parameters")

        # Get API version
        api_version = self.kwargs.get("api_version", "") or os.getenv("AZURE_OPENAI_API_VERSION", "2025-01-01-preview")

        # Get endpoint
        azure_endpoint = self.base_url
        if not azure_endpoint:
            azure_endpoint = os.getenv("AZURE_OPENAI_ENDPOINT", "")
            if not azure_endpoint:
                raise ValueError(
                    "Azure OpenAI endpoint not found, please set AZURE_OPENAI_ENDPOINT environment variable or provide it in the parameters")

        return AzureChatOpenAI(
            model=self.model_name or "gpt-4o",
            temperature=self.kwargs.get("temperature", 0.0),
            api_version=api_version,
            azure_endpoint=azure_endpoint,
            api_key=api_key
        )



import abc
import asyncio
from abc import abstractmethod
from dataclasses import field, dataclass
from typing import AsyncIterator, Any, Union, Iterator

from aworld.logs.util import logger
from aworld.output import Output
from aworld.output.base import RUN_FINISHED_SIGNAL


@dataclass
class Outputs(abc.ABC):
    """Base class for managing output streams in the AWorld framework.
    Provides abstract methods for adding and streaming outputs both synchronously and asynchronously.
    reference: https://github.com/openai/openai-agents-python/blob/main/src/agents/result.py
    """
    _metadata: dict = field(default_factory=dict)

    @abstractmethod
    async def add_output(self, output: Output):
        """Add an output asynchronously to the output stream.
        
        Args:
            output (Output): The output to be added
        """
        pass

    @abstractmethod
    def sync_add_output(self, output: Output):
        """Add an output synchronously to the output stream.
        
        Args:
            output (Output): The output to be added
        """
        pass

    @abstractmethod
    async def stream_events(self) -> Union[AsyncIterator[Output], list]:
        """Stream outputs asynchronously.
        
        Returns:
            AsyncIterator[Output]: An async iterator of outputs
        """
        pass

    @abstractmethod
    def sync_stream_events(self) -> Union[Iterator[Output], list]:
        """Stream outputs synchronously.
        
        Returns:
            Iterator[Output]: An iterator of outputs
        """
        pass

    @abstractmethod
    async def mark_completed(self):
        pass

    async def get_metadata(self) -> dict:
        return self._metadata

    async def set_metadata(self, metadata: dict):
        self._metadata = metadata

@dataclass
class AsyncOutputs(Outputs):
    """Intermediate class that implements the Outputs interface with async support.
    This class serves as a base for more specific async output implementations."""

    async def add_output(self, output: Output):
        pass

    def sync_add_output(self, output: Output):
        pass

    async def stream_events(self) -> Union[AsyncIterator[Output], list]:
        pass

    def sync_stream_events(self) -> Union[Iterator[Output]]:
        pass


@dataclass
class DefaultOutputs(Outputs):
    """DefaultAsyncOutputs """

    _outputs: list = field(default_factory=list)

    async def add_output(self, output: Output):
        self._outputs.append(output)

    def sync_add_output(self, output: Output):
        self._outputs.append(output)

    async def stream_events(self) -> Union[AsyncIterator[Output], list]:
        return self._outputs

    def sync_stream_events(self) -> Union[Iterator[Output], list]:
        return self._outputs

    async def mark_completed(self):
        pass


@dataclass
class StreamingOutputs(AsyncOutputs):
    """Concrete implementation of AsyncOutputs that provides streaming functionality.
    Manages a queue of outputs and handles streaming with error checking and task management."""

    # Task and input related fields
    # task: Task = Field(default=None)  # The task associated with these outputs
    input: Any = field(default=None)  # Input data for the task
    usage: dict = field(default=None)  # Usage statistics

    # State tracking
    is_complete: bool = field(default=False)  # Flag indicating if streaming is complete

    # Queue for managing outputs
    _output_queue: asyncio.Queue[Output] = field(
        default_factory=asyncio.Queue, repr=False
    )

    # Internal state management
    _visited_outputs: list[Output] = field(default_factory=list)
    _stored_exception: Exception | None = field(default=None, repr=False)  # Stores any exceptions that occur
    _run_impl_task: asyncio.Task[Any] | None = field(default=None, repr=False)  # The running task

    async def add_output(self, output: Output):
        """Add an output to the queue asynchronously.
        
        Args:
            output (Output): The output to be added to the queue
        """
        self._output_queue.put_nowait(output)

    async def stream_events(self) -> AsyncIterator[Output]:
        """Stream outputs asynchronously, handling cached outputs and new outputs from the queue.
        Includes error checking and task cleanup.
        
        Yields:
            Output: The next output in the stream
            
        Raises:
            Exception: Any stored exception that occurred during streaming
        """
        # First yield any cached outputs
        for output in self._visited_outputs:
            if output == RUN_FINISHED_SIGNAL:
                self._output_queue.task_done()
                return
            yield output

        # Main streaming loop
        while True:
            self._check_errors()
            if self._stored_exception:
                logger.debug("Breaking due to stored exception")
                self.is_complete = True
                break

            if self.is_complete and self._output_queue.empty():
                break

            try:
                output = await self._output_queue.get()
                self._visited_outputs.append(output)

            except asyncio.CancelledError:
                break

            if output == RUN_FINISHED_SIGNAL:
                self._output_queue.task_done()
                self._check_errors()
                break

            yield output
            self._output_queue.task_done()

        self._cleanup_tasks()

        if self._stored_exception:
            raise self._stored_exception

    def _check_errors(self):
        """Check for errors in the streaming process.
        Verifies step count and checks for exceptions in the running task.
        """
        # Check the task for any exceptions
        if self._run_impl_task and self._run_impl_task.done():
            exc = self._run_impl_task.exception()
            if exc and isinstance(exc, Exception):
                self._stored_exception = exc

    def _cleanup_tasks(self):
        """Clean up any running tasks by cancelling them if they're not done."""
        if self._run_impl_task and not self._run_impl_task.done():
            self._run_impl_task.cancel()

    async def mark_completed(self) -> None:
        """Mark the streaming process as completed by adding a RUN_FINISHED_SIGNAL to the queue."""
        await self._output_queue.put(RUN_FINISHED_SIGNAL)



from typing import (
    List,
    Dict,
    Union,
    Generator,
    AsyncGenerator,
)
from aworld.config import ConfigDict
from aworld.config.conf import AgentConfig, ClientType
from aworld.logs.util import logger
import aworld.trace as trace

from aworld.core.llm_provider_base import LLMProviderBase
from aworld.models.openai_provider import OpenAIProvider, AzureOpenAIProvider
from aworld.models.anthropic_provider import AnthropicProvider
from aworld.models.ant_provider import AntProvider
from aworld.models.model_response import ModelResponse

# Predefined model names for common providers
MODEL_NAMES = {
    "anthropic": ["claude-3-5-sonnet-20241022", "claude-3-5-sonnet-20240620", "claude-3-opus-20240229"],
    "openai": ["gpt-4o", "gpt-4", "gpt-3.5-turbo", "o3-mini", "gpt-4o-mini"],
    "azure_openai": ["gpt-4", "gpt-4-turbo", "gpt-4o", "gpt-35-turbo"],
}

# Endpoint patterns for identifying providers
ENDPOINT_PATTERNS = {
    "openai": ["api.openai.com"],
    "anthropic": ["api.anthropic.com", "claude-api"],
    "azure_openai": ["openai.azure.com"],
    "ant": ["zdfmng.alipay.com"],
}

# Provider class mapping
PROVIDER_CLASSES = {
    "openai": OpenAIProvider,
    "anthropic": AnthropicProvider,
    "azure_openai": AzureOpenAIProvider,
    "ant": AntProvider,
}


class LLMModel:
    """Unified large model interface, encapsulates different model implementations, provides a unified completion method.
    """

    def __init__(self, conf: Union[ConfigDict, AgentConfig] = None, custom_provider: LLMProviderBase = None, **kwargs):
        """Initialize unified model interface.

        Args:
            conf: Agent configuration, if provided, create model based on configuration.
            custom_provider: Custom LLMProviderBase instance, if provided, use it directly.
            **kwargs: Other parameters, may include:
                - base_url: Specify model endpoint.
                - api_key: API key.
                - model_name: Model name.
                - temperature: Temperature parameter.
        """

        # If custom_provider instance is provided, use it directly
        if custom_provider is not None:
            if not isinstance(custom_provider, LLMProviderBase):
                raise TypeError(
                    "custom_provider must be an instance of LLMProviderBase")
            self.provider_name = "custom"
            self.provider = custom_provider
            return

        # Get basic parameters
        base_url = kwargs.get("base_url") or (
            conf.llm_base_url if conf else None)
        model_name = kwargs.get("model_name") or (
            conf.llm_model_name if conf else None)
        llm_provider = conf.llm_provider if conf_contains_key(
            conf, "llm_provider") else None

        # Get API key from configuration (if any)
        if conf and conf.llm_api_key:
            kwargs["api_key"] = conf.llm_api_key

        # Identify provider
        self.provider_name = self._identify_provider(
            llm_provider, base_url, model_name)

        # Fill basic parameters
        kwargs['base_url'] = base_url
        kwargs['model_name'] = model_name

        # Fill parameters for llm provider
        kwargs['sync_enabled'] = conf.llm_sync_enabled if conf_contains_key(
            conf, "llm_sync_enabled") else True
        kwargs['async_enabled'] = conf.llm_async_enabled if conf_contains_key(
            conf, "llm_async_enabled") else True
        kwargs['client_type'] = conf.llm_client_type if conf_contains_key(
            conf, "llm_client_type") else ClientType.SDK

        kwargs.update(self._transfer_conf_to_args(conf))

        # Create model provider based on provider_name
        self._create_provider(**kwargs)

    def _transfer_conf_to_args(self, conf: Union[ConfigDict, AgentConfig] = None) -> dict:
        """
        Transfer parameters from conf to args

        Args:
            conf: config object
        """
        if not conf:
            return {}

        # Get all parameters from conf
        if type(conf).__name__ == 'AgentConfig':
            conf_dict = conf.model_dump()
        else:  # ConfigDict
            conf_dict = conf

        ignored_keys = ["llm_provider", "llm_base_url", "llm_model_name", "llm_api_key", "llm_sync_enabled",
                        "llm_async_enabled", "llm_client_type"]
        args = {}
        # Filter out used parameters and add remaining parameters to args
        for key, value in conf_dict.items():
            if key not in ignored_keys and value is not None:
                args[key] = value

        return args

    def _identify_provider(self, provider: str = None, base_url: str = None, model_name: str = None) -> str:
        """Identify LLM provider.

        Identification logic:
        1. If provider is specified and doesn't need to be overridden, use the specified provider.
        2. If base_url is provided, try to identify provider based on base_url.
        3. If model_name is provided, try to identify provider based on model_name.
        4. If none can be identified, default to "openai".

        Args:
            provider: Specified provider.
            base_url: Service URL.
            model_name: Model name.

        Returns:
            str: Identified provider.
        """
        # Default provider
        identified_provider = "openai"

        # Identify provider based on base_url
        if base_url:
            for p, patterns in ENDPOINT_PATTERNS.items():
                if any(pattern in base_url for pattern in patterns):
                    identified_provider = p
                    logger.info(
                        f"Identified provider: {identified_provider} based on base_url: {base_url}")
                    return identified_provider

        # Identify provider based on model_name
        if model_name and not base_url:
            for p, models in MODEL_NAMES.items():
                if model_name in models or any(model_name.startswith(model) for model in models):
                    identified_provider = p
                    logger.info(
                        f"Identified provider: {identified_provider} based on model_name: {model_name}")
                    break

        if provider and provider in PROVIDER_CLASSES and identified_provider and identified_provider != provider:
            logger.warning(
                f"Provider mismatch: {provider} != {identified_provider}, using {provider} as provider")
            identified_provider = provider

        return identified_provider

    def _create_provider(self, **kwargs):
        """Return the corresponding provider instance based on provider.

        Args:
            **kwargs: Parameters, may include:
                - base_url: Model endpoint.
                - api_key: API key.
                - model_name: Model name.
                - temperature: Temperature parameter.
                - timeout: Timeout.
                - max_retries: Maximum number of retries.
        """
        self.provider = PROVIDER_CLASSES[self.provider_name](**kwargs)

    @classmethod
    def supported_providers(cls) -> list[str]:
        return list(PROVIDER_CLASSES.keys())

    def supported_models(self) -> list[str]:
        """Get supported models for the current provider.
        Returns:
            list: Supported models.
        """
        return self.provider.supported_models() if self.provider else []

    async def acompletion(self,
                          messages: List[Dict[str, str]],
                          temperature: float = 0.0,
                          max_tokens: int = None,
                          stop: List[str] = None,
                          **kwargs) -> ModelResponse:
        """Asynchronously call model to generate response.

        Args:
            messages: Message list, format is [{"role": "system", "content": "..."}, {"role": "user", "content": "..."}].
            temperature: Temperature parameter.
            max_tokens: Maximum number of tokens to generate.
            stop: List of stop sequences.
            **kwargs: Other parameters.

        Returns:
            ModelResponse: Unified model response object.
        """
        # Call provider's acompletion method directly
        return await self.provider.acompletion(
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            **kwargs
        )

    def completion(self,
                   messages: List[Dict[str, str]],
                   temperature: float = 0.0,
                   max_tokens: int = None,
                   stop: List[str] = None,
                   **kwargs) -> ModelResponse:
        """Synchronously call model to generate response.

        Args:
            messages: Message list, format is [{"role": "system", "content": "..."}, {"role": "user", "content": "..."}].
            temperature: Temperature parameter.
            max_tokens: Maximum number of tokens to generate.
            stop: List of stop sequences.
            **kwargs: Other parameters.

        Returns:
            ModelResponse: Unified model response object.
        """
        # Call provider's completion method directly
        return self.provider.completion(
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            **kwargs
        )

    def stream_completion(self,
                          messages: List[Dict[str, str]],
                          temperature: float = 0.0,
                          max_tokens: int = None,
                          stop: List[str] = None,
                          **kwargs) -> Generator[ModelResponse, None, None]:
        """Synchronously call model to generate streaming response.

        Args:
            messages: Message list, format is [{"role": "system", "content": "..."}, {"role": "user", "content": "..."}].
            temperature: Temperature parameter.
            max_tokens: Maximum number of tokens to generate.
            stop: List of stop sequences.
            **kwargs: Other parameters.

        Returns:
            Generator yielding ModelResponse chunks.
        """
        # Call provider's stream_completion method directly
        return self.provider.stream_completion(
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            **kwargs
        )

    async def astream_completion(self,
                                 messages: List[Dict[str, str]],
                                 temperature: float = 0.0,
                                 max_tokens: int = None,
                                 stop: List[str] = None,
                                 **kwargs) -> AsyncGenerator[ModelResponse, None]:
        """Asynchronously call model to generate streaming response.

        Args:
            messages: Message list, format is [{"role": "system", "content": "..."}, {"role": "user", "content": "..."}].
            temperature: Temperature parameter.
            max_tokens: Maximum number of tokens to generate.
            stop: List of stop sequences.
            **kwargs: Other parameters, may include:
                - base_url: Specify model endpoint.
                - api_key: API key.
                - model_name: Model name.

        Returns:
            AsyncGenerator yielding ModelResponse chunks.
        """
        # Call provider's astream_completion method directly
        async for chunk in self.provider.astream_completion(
                messages=messages,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                **kwargs
        ):
            yield chunk


def register_llm_provider(provider: str, provider_class: type):
    """Register a custom LLM provider.

    Args:
        provider: Provider name.
        provider_class: Provider class, must inherit from LLMProviderBase.
    """
    if not issubclass(provider_class, LLMProviderBase):
        raise TypeError("provider_class must be a subclass of LLMProviderBase")
    PROVIDER_CLASSES[provider] = provider_class


def conf_contains_key(conf: Union[ConfigDict, AgentConfig], key: str) -> bool:
    """Check if conf contains key.
    Args:
        conf: Config object.
        key: Key to check.
    Returns:
        bool: Whether conf contains key.
    """
    if not conf:
        return False
    if type(conf).__name__ == 'AgentConfig':
        return hasattr(conf, key)
    else:
        return key in conf


def get_llm_model(conf: Union[ConfigDict, AgentConfig] = None,
                  custom_provider: LLMProviderBase = None,
                  **kwargs) -> Union[LLMModel, 'ChatOpenAI']:
    """Get a unified LLM model instance.

    Args:
        conf: Agent configuration, if provided, create model based on configuration.
        custom_provider: Custom LLMProviderBase instance, if provided, use it directly.
        **kwargs: Other parameters, may include:
            - base_url: Specify model endpoint.
            - api_key: API key.
            - model_name: Model name.
            - temperature: Temperature parameter.

    Returns:
        Unified model interface.
    """
    # Create and return LLMModel instance directly
    llm_provider = conf.llm_provider if conf_contains_key(
        conf, "llm_provider") else None

    if (llm_provider == "chatopenai"):
        from langchain_openai import ChatOpenAI

        base_url = kwargs.get("base_url") or (conf.llm_base_url if conf_contains_key(conf, "llm_base_url") else None)
        model_name = kwargs.get("model_name") or (conf.llm_model_name if conf_contains_key(conf, "llm_model_name") else None)
        api_key = kwargs.get("api_key") or (conf.llm_api_key if conf_contains_key(conf, "llm_api_key") else None)

        return ChatOpenAI(
            model=model_name,
            temperature=kwargs.get("temperature", conf.llm_temperature if conf_contains_key(
                conf, "llm_temperature") else 0.0),
            base_url=base_url,
            api_key=api_key,
        )

    return LLMModel(conf=conf, custom_provider=custom_provider, **kwargs)


def call_llm_model(
        llm_model: LLMModel,
        messages: List[Dict[str, str]],
        temperature: float = 0.0,
        max_tokens: int = None,
        stop: List[str] = None,
        stream: bool = False,
        **kwargs
) -> Union[ModelResponse, Generator[ModelResponse, None, None]]:
    """Convenience function to call LLM model.

    Args:
        llm_model: LLM model instance.
        messages: Message list.
        temperature: Temperature parameter.
        max_tokens: Maximum number of tokens to generate.
        stop: List of stop sequences.
        stream: Whether to return a streaming response.
        **kwargs: Other parameters.

    Returns:
        Model response or response generator.
    """
    with trace.span(llm_model.provider.model_name, run_type=trace.RunType.LLM) as llm_span:
        if stream:
            return llm_model.stream_completion(
                messages=messages,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                **kwargs
            )
        else:
            return llm_model.completion(
                messages=messages,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                **kwargs
            )


async def acall_llm_model(
        llm_model: LLMModel,
        messages: List[Dict[str, str]],
        temperature: float = 0.0,
        max_tokens: int = None,
        stop: List[str] = None,
        stream: bool = False,
        **kwargs
) -> Union[ModelResponse, AsyncGenerator[ModelResponse, None]]:
    """Convenience function to asynchronously call LLM model.

    Args:
        llm_model: LLM model instance.
        messages: Message list.
        temperature: Temperature parameter.
        max_tokens: Maximum number of tokens to generate.
        stop: List of stop sequences.
        stream: Whether to return a streaming response.
        **kwargs: Other parameters.

    Returns:
        Model response or response generator.
    """
    async with trace.span(llm_model.provider.model_name, run_type=trace.RunType.LLM) as llm_span:
        if stream:
            async def _stream_wrapper():
                async for chunk in llm_model.astream_completion(
                        messages=messages,
                        temperature=temperature,
                        max_tokens=max_tokens,
                        stop=stop,
                        **kwargs
                ):
                    yield chunk

            return _stream_wrapper()
        else:
            return await llm_model.acompletion(
                messages=messages,
                temperature=temperature,
                max_tokens=max_tokens,
                stop=stop,
                **kwargs
            )



from typing import Any, List, TypeVar, Union, Literal, TypedDict, Dict

DataRow = TypeVar('DataRow')


class BaseCondition(TypedDict):
    field: str
    value: Any
    op: Literal[
        'eq', 'ne', 'gt', 'gte', 'lt', 'lte',
        'in', 'not_in', 'like', 'not_like',
        'is_null', 'is_not_null'
    ]


class LogicalCondition(TypedDict):
    and_: List['QueryCondition']
    or_: List['QueryCondition']


QueryCondition = Union[BaseCondition, LogicalCondition]


class QueryBuilder:
    '''
    Query builder for replay buffer. result example:
    {
        "and": [
            {"field": "field1", "value": "value1", "op": "eq"}, 
            {"or": [{"field": "field2", "value": "value2", "op": "eq"}, {"field": "field3", "value": "value3", "op": "eq"}]}
        ]
    }
    '''

    def __init__(self) -> None:
        self.conditions: List[Dict[str, any]] = []
        self.logical_ops: List[str] = []

    def eq(self, field: str, value: any) -> 'QueryBuilder':
        self.conditions.append({"field": field, "value": value, "op": "eq"})
        return self

    def ne(self, field: str, value: any) -> 'QueryBuilder':
        self.conditions.append({"field": field, "value": value, "op": "ne"})
        return self

    def gt(self, field: str, value: any) -> 'QueryBuilder':
        self.conditions.append({"field": field, "value": value, "op": "gt"})
        return self

    def gte(self, field: str, value: any) -> 'QueryBuilder':
        self.conditions.append({"field": field, "value": value, "op": "gte"})
        return self

    def lt(self, field: str, value: any) -> 'QueryBuilder':
        self.conditions.append({"field": field, "value": value, "op": "lt"})
        return self

    def lte(self, field: str, value: any) -> 'QueryBuilder':
        self.conditions.append({"field": field, "value": value, "op": "lte"})
        return self

    def in_(self, field: str, value: any) -> 'QueryBuilder':
        self.conditions.append({"field": field, "value": value, "op": "in"})
        return self

    def not_in(self, field: str, value: any) -> 'QueryBuilder':
        self.conditions.append(
            {"field": field, "value": value, "op": "not_in"})
        return self

    def like(self, field: str, value: any) -> 'QueryBuilder':
        self.conditions.append({"field": field, "value": value, "op": "like"})
        return self

    def not_like(self, field: str, value: any) -> 'QueryBuilder':
        self.conditions.append(
            {"field": field, "value": value, "op": "not_like"})
        return self

    def is_null(self, field: str) -> 'QueryBuilder':
        self.conditions.append({"field": field, "op": "is_null"})
        return self

    def is_not_null(self, field: str) -> 'QueryBuilder':
        self.conditions.append({"field": field, "op": "is_not_null"})
        return self

    def and_(self) -> 'QueryBuilder':
        self.logical_ops.append("and_")
        return self

    def or_(self) -> 'QueryBuilder':
        self.logical_ops.append("or_")
        return self

    def nested(self, builder: 'QueryBuilder') -> 'QueryBuilder':
        self.conditions.append({"nested": builder.build()})
        return self

    def build(self) -> QueryCondition:
        conditions = self.conditions  # all conditions（including nested）
        operators = self.logical_ops

        # Validate condition and operator counts (n conditions need n-1 operators)
        if len(operators) != len(conditions) - 1:
            raise ValueError("Mismatch between condition and operator counts")

        # Use stack to handle operator precedence (simplified version supporting and/or)
        stack: List[Union[Dict[str, any], str]] = []

        for i, item in enumerate(conditions):
            if i == 0:
                # First element goes directly to stack (condition or nested)
                stack.append(item)
                continue

            # Pop stack top as left operand
            left = stack.pop()
            op = operators[i-1]       # Current operator (and/or)
            right = item              # Right operand (current condition)

            # Build logical expression: {op: [left, right]}
            expr = {op: [left, right]}
            # Push result back to stack for further operations
            stack.append(expr)

        # Process nested conditions (recursive unfolding)
        def process_nested(cond: any) -> any:
            if isinstance(cond, dict):
                if "nested" in cond:
                    # Recursively process sub-conditions
                    return process_nested(cond["nested"])
                # Recursively process child elements
                return {k: process_nested(v) for k, v in cond.items()}
            elif isinstance(cond, list):
                return [process_nested(item) for item in cond]
            return cond

        # Final result: only one element left in stack, return after processing nested
        result = stack[0] if stack else None
        return process_nested(result) if result else None


class QueryFilter:
    '''
    Query filter for replay buffer.
    '''

    def __init__(self, query_condition: QueryCondition) -> None:
        self.query_condition = query_condition

    def _get_field_value(self, row: DataRow, field: str) -> Any:
        '''
        Get field value from row.
        '''
        obj = row
        for part in field.split('.'):
            obj = getattr(obj, part, None)
            if obj is None:
                break
        return obj

    def _do_check(self, row: DataRow, condition: QueryCondition) -> bool:
        """
        check if row match condition
        """
        if condition is None:
            return True
        if "field" in condition and "op" in condition:
            field_val = self._get_field_value(row, condition["field"])
            op = condition["op"]
            target_val = condition["value"]

            if op == "eq":
                return field_val == target_val
            if op == "ne":
                return field_val != target_val
            if op == "gt":
                return field_val > target_val
            if op == "gte":
                return field_val >= target_val
            if op == "lt":
                return field_val < target_val
            if op == "lte":
                return field_val <= target_val
            if op == "in":
                return field_val in target_val
            if op == "not_in":
                return field_val not in target_val
            if op == "like":
                return target_val in field_val
            if op == "not_like":
                return target_val not in field_val
            if op == "is_null":
                return field_val is None
            if op == "is_not_null":
                return field_val is not None

            return False

        elif "and_" in condition or "or_" in condition:
            if "and_" in condition:
                return all(self._do_check(row, c) for c in condition["and_"])
            if "or_" in condition:
                return any(self._do_check(row, c) for c in condition["or_"])
            return False

        return False

    def check_condition(self, row: DataRow) -> bool:
        """
        check if row match condition
        """
        return self._do_check(row, self.query_condition)

    def filter(self, rows: List[DataRow]) -> List[DataRow]:
        """filter rows by condition
        Args:
            rows (List[DataRow]): List of rows to filter.
            query_condition (QueryCondition): Query condition.
        Returns:
            List[DataRow]: List of rows that match the condition.
        """
        condition = self.query_condition
        if not condition:
            return rows
        return [row for row in rows if self.check_condition(row)]



# coding: utf-8
# Copyright (c) 2025 inclusionAI.
import os
import traceback
from typing import Sequence, Union
from aworld.core.common import ActionModel
from aworld.trace.context_manager import TraceManager, trace_configure
from aworld.trace.constants import RunType
from aworld.logs.log import set_log_provider, instrument_logging
from aworld.logs.util import logger, trace_logger

if os.getenv("LOGFIRE_WRITE_TOKEN"):
    trace_configure(
        backends=["logfire"],
        write_token=os.getenv("LOGFIRE_WRITE_TOKEN")
    )
    set_log_provider(provider="otlp", backend="logfire",
                     write_token=os.getenv("LOGFIRE_WRITE_TOKEN"))
elif os.getenv("OTLP_TRACES_ENDPOINT"):
    trace_configure(
        backends=["other_otlp"]
    )
else:
    logger.warning("LOGFIRE_WRITE_TOKEN is not set, using memory backend")
    trace_configure(
        backends=["memory"]
    )

instrument_logging(trace_logger)


def get_tool_name(tool_name: str,
                  action: Union[ActionModel, Sequence[ActionModel]]) -> tuple[str, RunType]:
    if tool_name == "mcp" and action:
        try:
            if isinstance(action, (list, tuple)):
                action = action[0]
            mcp_name = action.action_name.split("__")[0]
            return (mcp_name, RunType.MCP)
        except ValueError:
            logger.warning(traceback.format_exc())
            return (tool_name, RunType.MCP)
    return (tool_name, RunType.TOOL)


GLOBAL_TRACE_MANAGER: TraceManager = TraceManager()
span = GLOBAL_TRACE_MANAGER.span
func_span = GLOBAL_TRACE_MANAGER.func_span
auto_tracing = GLOBAL_TRACE_MANAGER.auto_tracing
get_current_span = GLOBAL_TRACE_MANAGER.get_current_span
new_manager = GLOBAL_TRACE_MANAGER.get_current_span



# coding: utf-8
"""
oss.py
Utility class for OSS (Object Storage Service) operations.
Provides simple methods for data operations: upload, read, delete, update.
"""
import os
import json
import tempfile
from typing import Optional, Dict, List, Any, Tuple, Union, BinaryIO, TextIO, IO, AnyStr

from aworld.logs.util import logger


class OSSClient:
    """
    A utility class for OSS (Object Storage Service) operations.
    Provides methods for data operations: upload, read, delete, update.
    """
    
    def __init__(self, 
                 access_key_id: Optional[str] = None, 
                 access_key_secret: Optional[str] = None, 
                 endpoint: Optional[str] = None, 
                 bucket_name: Optional[str] = None,
                 enable_export: Optional[bool] = None):
        """
        Initialize OSSClient with credentials.
        
        Args:
            access_key_id: OSS access key ID. If None, will try to get from environment variable OSS_ACCESS_KEY_ID
            access_key_secret: OSS access key secret. If None, will try to get from environment variable OSS_ACCESS_KEY_SECRET
            endpoint: OSS endpoint. If None, will try to get from environment variable OSS_ENDPOINT
            bucket_name: OSS bucket name. If None, will try to get from environment variable OSS_BUCKET_NAME
            enable_export: Whether to enable OSS export. If None, will try to get from environment variable EXPORT_REPLAY_TRACE_TO_OSS
        """
        self.access_key_id = access_key_id or os.getenv('OSS_ACCESS_KEY_ID')
        self.access_key_secret = access_key_secret or os.getenv('OSS_ACCESS_KEY_SECRET')
        self.endpoint = endpoint or os.getenv('OSS_ENDPOINT')
        self.bucket_name = bucket_name or os.getenv('OSS_BUCKET_NAME')
        self.enable_export = enable_export if enable_export is not None else os.getenv("EXPORT_REPLAY_TRACE_TO_OSS", "false").lower() == "true"
        self.bucket = None
        self._initialized = False
        
    def initialize(self) -> bool:
        """
        Initialize the OSS client with the provided or environment credentials.
        
        Returns:
            bool: True if initialization is successful, False otherwise
        """
        if self._initialized:
            return True
            
        if not self.enable_export:
            logger.info("OSS export is disabled. Set EXPORT_REPLAY_TRACE_TO_OSS=true to enable.")
            return False
            
        if not all([self.access_key_id, self.access_key_secret, self.endpoint, self.bucket_name]):
            logger.warn("Missing required OSS credentials. Please provide all required parameters or set environment variables.")
            return False
            
        try:
            import oss2
            auth = oss2.Auth(self.access_key_id, self.access_key_secret)
            self.bucket = oss2.Bucket(auth, self.endpoint, self.bucket_name)
            self._initialized = True
            return True
        except ImportError:
            logger.warn("Failed to import oss2 module. Please install it with 'pip install oss2'.")
            return False
        except Exception as e:
            logger.warn(f"Failed to initialize OSS client. Error: {str(e)}")
            return False

    # ---- Basic Data Operation Methods ----
    
    def upload_data(self, data: Union[IO[AnyStr], str, bytes, dict], oss_key: str) -> bool:
        """
        Upload data to OSS. Supports various types of data:
        - In-memory file objects (IO[AnyStr])
        - Strings (str)
        - Bytes (bytes)
        - Dictionaries (dict), will be automatically converted to JSON
        - File paths (str)
        
        Args:
            data: Data to upload, can be a file object or other supported types
            oss_key: The key (path) in OSS where the data will be stored
            
        Returns:
            bool: True if successful, False otherwise
        """
        if not self.initialize():
            logger.warn("OSS client not initialized or export is disabled")
            return False
            
        try:
            # Handle file objects
            if hasattr(data, 'read'):
                content = data.read()
                if isinstance(content, str):
                    content = content.encode('utf-8')
                self.bucket.put_object(oss_key, content)
                logger.info(f"Successfully uploaded memory file to OSS: {oss_key}")
                return True
                
            # Handle dictionaries
            if isinstance(data, dict):
                content = json.dumps(data, ensure_ascii=False).encode('utf-8')
                self.bucket.put_object(oss_key, content)
                return True
                
            # Handle strings
            if isinstance(data, str):
                # Check if it's a file path
                if os.path.isfile(data):
                    self.bucket.put_object_from_file(oss_key, data)
                    logger.info(f"Successfully uploaded file {data} to OSS: {oss_key}")
                    return True
                # Otherwise treat as string content
                content = data.encode('utf-8')
                self.bucket.put_object(oss_key, content)
                return True
                
            # Handle bytes
            self.bucket.put_object(oss_key, data)
            logger.info(f"Successfully uploaded data to OSS: {oss_key}")
            return True
        except Exception as e:
            logger.warn(f"Failed to upload data to OSS: {str(e)}")
            return False
    
    def read_data(self, oss_key: str, as_json: bool = False) -> Union[bytes, dict, str, None]:
        """
        Read data from OSS.
        
        Args:
            oss_key: The key (path) in OSS of the data to read
            as_json: If True, parse the data as JSON and return a dict
            
        Returns:
            The data as bytes, dict (if as_json=True), or None if failed
        """
        if not self.initialize():
            logger.warn("OSS client not initialized or export is disabled")
            return None
            
        try:
            # Read data
            result = self.bucket.get_object(oss_key)
            data = result.read()
            
            # Convert to string or JSON if requested
            if as_json:
                return json.loads(data)
            
            return data
        except Exception as e:
            logger.warn(f"Failed to read data from OSS: {str(e)}")
            return None
    
    def read_text(self, oss_key: str) -> Optional[str]:
        """
        Read text data from OSS.
        
        Args:
            oss_key: The key (path) in OSS of the text to read
            
        Returns:
            str: The text data, or None if failed
        """
        data = self.read_data(oss_key)
        if data is not None:
            try:
                return data.decode('utf-8')
            except Exception as e:
                logger.warn(f"Failed to decode data as UTF-8: {str(e)}")
        return None
    
    def delete_data(self, oss_key: str) -> bool:
        """
        Delete data from OSS.
        
        Args:
            oss_key: The key (path) in OSS of the data to delete
            
        Returns:
            bool: True if successful, False otherwise
        """
        if not self.initialize():
            logger.warn("OSS client not initialized or export is disabled")
            return False
            
        try:
            self.bucket.delete_object(oss_key)
            logger.info(f"Successfully deleted data from OSS: {oss_key}")
            return True
        except Exception as e:
            logger.warn(f"Failed to delete data from OSS: {str(e)}")
            return False
    
    def update_data(self, oss_key: str, data: Union[IO[AnyStr], str, bytes, dict]) -> bool:
        """
        Update data in OSS (delete and upload).
        
        Args:
            oss_key: The key (path) in OSS of the data to update
            data: New data to upload, can be a file object or other supported types
            
        Returns:
            bool: True if successful, False otherwise
        """
        # For OSS, update is the same as upload (it overwrites)
        return self.upload_data(data, oss_key)
    
    def update_json(self, oss_key: str, update_dict: dict) -> bool:
        """
        Update JSON data in OSS by merging with existing data.
        
        Args:
            oss_key: The key (path) in OSS of the JSON data to update
            update_dict: Dictionary with fields to update
            
        Returns:
            bool: True if successful, False otherwise
        """
        if not self.initialize():
            logger.warn("OSS client not initialized or export is disabled")
            return False
            
        try:
            # Read existing data
            existing_data = self.read_data(oss_key, as_json=True)
            if existing_data is None:
                existing_data = {}
                
            # Update data
            if isinstance(existing_data, dict):
                existing_data.update(update_dict)
            else:
                logger.warn(f"Existing data is not a dictionary: {oss_key}")
                return False
                
            # Upload updated data
            return self.upload_data(existing_data, oss_key)
        except Exception as e:
            logger.warn(f"Failed to update JSON data in OSS: {str(e)}")
            return False
    
    # ---- File Operation Methods ----
    
    def upload_file(self, local_file: str, oss_key: Optional[str] = None) -> bool:
        """
        Upload a local file to OSS.
        
        Args:
            local_file: Path to the local file
            oss_key: The key (path) in OSS where the file will be stored. 
                     If None, will use the basename of the local file
                     
        Returns:
            bool: True if successful, False otherwise
        """
        if not self.initialize():
            logger.warn("OSS client not initialized or export is disabled")
            return False
            
        try:
            if not os.path.exists(local_file):
                logger.warn(f"Local file {local_file} does not exist")
                return False
                
            if oss_key is None:
                oss_key = f"uploads/{os.path.basename(local_file)}"
                
            self.bucket.put_object_from_file(oss_key, local_file)
            logger.info(f"Successfully uploaded {local_file} to OSS: {oss_key}")
            return True
        except Exception as e:
            logger.warn(f"Failed to upload {local_file} to OSS: {str(e)}")
            return False
    
    def download_file(self, oss_key: str, local_file: str) -> bool:
        """
        Download a file from OSS to local.
        
        Args:
            oss_key: The key (path) in OSS of the file to download
            local_file: Path where the downloaded file will be saved
            
        Returns:
            bool: True if successful, False otherwise
        """
        if not self.initialize():
            logger.warn("OSS client not initialized or export is disabled")
            return False
            
        try:
            # Ensure the directory exists
            os.makedirs(os.path.dirname(os.path.abspath(local_file)), exist_ok=True)
            
            # Download the file
            self.bucket.get_object_to_file(oss_key, local_file)
            logger.info(f"Successfully downloaded {oss_key} to {local_file}")
            return True
        except Exception as e:
            logger.warn(f"Failed to download {oss_key} from OSS: {str(e)}")
            return False
    
    def list_objects(self, prefix: str = "", delimiter: str = "") -> List[Dict[str, Any]]:
        """
        List objects in the OSS bucket with the given prefix.
        
        Args:
            prefix: Prefix to filter objects
            delimiter: Delimiter for hierarchical listing
            
        Returns:
            List of objects with their properties
        """
        if not self.initialize():
            logger.warn("OSS client not initialized or export is disabled")
            return []
            
        try:
            result = []
            for obj in self.bucket.list_objects(prefix=prefix, delimiter=delimiter).object_list:
                result.append({
                    'key': obj.key,
                    'size': obj.size,
                    'last_modified': obj.last_modified
                })
            return result
        except Exception as e:
            logger.warn(f"Failed to list objects with prefix {prefix}: {str(e)}")
            return []
    
    # ---- Advanced Operation Methods ----
    
    def exists(self, oss_key: str) -> bool:
        """
        Check if an object exists in OSS.
        
        Args:
            oss_key: The key (path) in OSS to check
            
        Returns:
            bool: True if the object exists, False otherwise
        """
        if not self.initialize():
            logger.warn("OSS client not initialized or export is disabled")
            return False
            
        try:
            # Use head_object to check if the object exists
            self.bucket.head_object(oss_key)
            return True
        except:
            return False
    
    def copy_object(self, source_key: str, target_key: str) -> bool:
        """
        Copy an object within the same bucket.
        
        Args:
            source_key: The source object key
            target_key: The target object key
            
        Returns:
            bool: True if successful, False otherwise
        """
        if not self.initialize():
            logger.warn("OSS client not initialized or export is disabled")
            return False
            
        try:
            self.bucket.copy_object(self.bucket_name, source_key, target_key)
            logger.info(f"Successfully copied {source_key} to {target_key}")
            return True
        except Exception as e:
            logger.warn(f"Failed to copy {source_key} to {target_key}: {str(e)}")
            return False
    
    def get_object_url(self, oss_key: str, expires: int = 3600) -> Optional[str]:
        """
        Generate a temporary URL for accessing an object.
        
        Args:
            oss_key: The key (path) in OSS of the object
            expires: URL expiration time in seconds
            
        Returns:
            str: The signed URL, or None if failed
        """
        if not self.initialize():
            logger.warn("OSS client not initialized or export is disabled")
            return None
            
        try:
            url = self.bucket.sign_url('GET', oss_key, expires)
            return url
        except Exception as e:
            logger.warn(f"Failed to generate URL for {oss_key}: {str(e)}")
            return None
    
    def upload_directory(self, local_dir: str, oss_prefix: str = "") -> Tuple[bool, List[str]]:
        """
        Upload an entire directory to OSS.
        
        Args:
            local_dir: Path to the local directory
            oss_prefix: Prefix to prepend to all uploaded files
            
        Returns:
            Tuple of (success, list of uploaded files)
        """
        if not self.initialize():
            logger.warn("OSS client not initialized or export is disabled")
            return False, []
            
        if not os.path.isdir(local_dir):
            logger.warn(f"Local directory {local_dir} does not exist or is not a directory")
            return False, []
            
        uploaded_files = []
        errors = []
        
        for root, _, files in os.walk(local_dir):
            for file in files:
                local_file = os.path.join(root, file)
                rel_path = os.path.relpath(local_file, local_dir)
                oss_key = os.path.join(oss_prefix, rel_path).replace("\\", "/")
                
                success = self.upload_file(local_file, oss_key)
                if success:
                    uploaded_files.append(oss_key)
                else:
                    errors.append(local_file)
        
        if errors:
            logger.warn(f"Failed to upload {len(errors)} files")
            return False, uploaded_files
        return True, uploaded_files


def get_oss_client(access_key_id: Optional[str] = None, 
                  access_key_secret: Optional[str] = None, 
                  endpoint: Optional[str] = None, 
                  bucket_name: Optional[str] = None,
                  enable_export: Optional[bool] = None) -> OSSClient:
    """
    Factory function to create and initialize an OSSClient.
    
    Args:
        access_key_id: OSS access key ID
        access_key_secret: OSS access key secret
        endpoint: OSS endpoint
        bucket_name: OSS bucket name
        enable_export: Whether to enable OSS export
        
    Returns:
        OSSClient: An initialized OSSClient instance
    """
    client = OSSClient(
        access_key_id=access_key_id, 
        access_key_secret=access_key_secret, 
        endpoint=endpoint, 
        bucket_name=bucket_name,
        enable_export=enable_export
    )
    client.initialize()
    return client


# ---- Test Cases ----
if __name__ == "__main__":
    """
    OSS tool class test cases
    Note: Before running the tests, you need to set the following environment variables,
    or provide the parameters directly in the test code:
    - OSS_ACCESS_KEY_ID
    - OSS_ACCESS_KEY_SECRET
    - OSS_ENDPOINT
    - OSS_BUCKET_NAME
    - EXPORT_REPLAY_TRACE_TO_OSS=true
    """
    import io
    import time
    
    # Test configuration
    TEST_PREFIX = f"test/oss_utils_123"  # Use timestamp to avoid conflicts
    
    # Initialize client
    # Method 1: Using environment variables
    # oss_client = get_oss_client(enable_export=True)
    
    # Method 2: Provide parameters directly
    oss_client = get_oss_client(
        access_key_id="",  # Replace with your actual access key ID
        access_key_secret="",  # Replace with your actual access key secret
        endpoint="",  # Replace with your actual OSS endpoint
        bucket_name="",  # Replace with your actual bucket name
        enable_export=True
    )

    # Test 1: Upload string data
    print("\nTest 1: Upload string data")
    text_key = f"{TEST_PREFIX}/text.txt"
    success = oss_client.upload_data("This is a test text", text_key)
    print(f"Upload string data: {'Success' if success else 'Failed'}")

    # Test 2: Upload dictionary data (automatically converted to JSON)
    print("\nTest 2: Upload dictionary data")
    json_key = f"{TEST_PREFIX}/data.json"
    data = {
        "name": "Test data",
        "values": [1, 2, 3],
        "nested": {
            "key": "value"
        }
    }
    success = oss_client.upload_data(data, json_key)
    print(f"Upload dictionary data: {'Success' if success else 'Failed'}")

    # Test 3: Upload in-memory binary file object
    print("\nTest 3: Upload in-memory binary file object")
    binary_key = f"{TEST_PREFIX}/binary.dat"
    binary_data = io.BytesIO(b"\x00\x01\x02\x03\x04")
    success = oss_client.upload_data(binary_data, binary_key)
    print(f"Upload binary file object: {'Success' if success else 'Failed'}")

    # Test 4: Upload in-memory text file object
    print("\nTest 4: Upload in-memory text file object")
    text_file_key = f"{TEST_PREFIX}/text_file.txt"
    text_file = io.StringIO("This is the content of an in-memory text file")
    success = oss_client.upload_data(text_file, text_file_key)
    print(f"Upload text file object: {'Success' if success else 'Failed'}")

    # Test 5: Create and upload temporary file
    print("\nTest 5: Create and upload temporary file")
    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        tmp.write(b"This is the content of a temporary file")
        tmp_path = tmp.name

    file_key = f"{TEST_PREFIX}/temp_file.txt"
    success = oss_client.upload_file(tmp_path, file_key)
    print(f"Upload temporary file: {'Success' if success else 'Failed'}")
    os.unlink(tmp_path)  # Delete temporary file

    # Test 6: Read text data
    print("\nTest 6: Read text data")
    content = oss_client.read_text(text_key)
    print(f"Read text data: {content}")

    # Test 7: Read JSON data
    print("\nTest 7: Read JSON data")
    json_content = oss_client.read_data(json_key, as_json=True)
    print(f"Read JSON data: {json_content}")

    # Test 8: Update JSON data (merge method)
    print("\nTest 8: Update JSON data")
    update_data = {"updated": True, "timestamp": time.time()}
    success = oss_client.update_json(json_key, update_data)
    print(f"Update JSON data: {'Success' if success else 'Failed'}")

    # View updated JSON data
    updated_json = oss_client.read_data(json_key, as_json=True)
    print(f"Updated JSON data: {updated_json}")

    # Test 9: Overwrite existing data
    print("\nTest 9: Overwrite existing data")
    success = oss_client.upload_data("This is the overwritten text", text_key)
    print(f"Overwrite existing data: {'Success' if success else 'Failed'}")

    # View overwritten data
    new_content = oss_client.read_text(text_key)
    print(f"Overwritten text data: {new_content}")

    # Test 10: List objects
    print("\nTest 10: List objects")
    objects = oss_client.list_objects(prefix=TEST_PREFIX)
    print(f"Found {len(objects)} objects:")
    for obj in objects:
        print(f"  - {obj['key']} (Size: {obj['size']} bytes, Modified: {obj['last_modified']})")

    # Test 11: Generate temporary URL
    print("\nTest 11: Generate temporary URL")
    url = oss_client.get_object_url(text_key, expires=300)  # 5 minutes expiration
    print(f"Temporary URL: {url}")

    # Test 12: Copy object
    print("\nTest 12: Copy object")
    copy_key = f"{TEST_PREFIX}/copy_of_text.txt"
    success = oss_client.copy_object(text_key, copy_key)
    print(f"Copy object: {'Success' if success else 'Failed'}")

    # Test 13: Check if object exists
    print("\nTest 13: Check if object exists")
    exists = oss_client.exists(text_key)
    print(f"Object {text_key} exists: {exists}")

    non_existent_key = f"{TEST_PREFIX}/non_existent.txt"
    exists = oss_client.exists(non_existent_key)
    print(f"Object {non_existent_key} exists: {exists}")

    # Test 14: Delete objects
    print("\nTest 14: Delete objects")
    for obj in objects:
        success = oss_client.delete_data(obj['key'])
        print(f"Delete object {obj['key']}: {'Success' if success else 'Failed'}")

    # Cleanup: Delete copied object (may not be included in the previous list)
    oss_client.delete_data(copy_key)
    
    print("\nTests completed!") 


# coding: utf-8
# Copyright (c) 2025 inclusionAI.

import os
import getpass
import time

__version__ = '0.2.4'

version_template = \
    """# auto generated
class VersionInfo:
    BUILD_DATE = "{BUILD_DATE}"
    BUILD_VERSION = "{BUILD_VERSION}"
    BUILD_USER = "ganrunsheng" 
    SCENARIO = "{SCENARIO}"
"""


def generate_version_info(directory_path: str = None, scenario: str = "", version: str = None):
    if directory_path is None:
        directory_path = os.path.dirname(__file__)
    with open(os.path.join(directory_path, "version_info.py"), "w") as f:
        version_info = _build_version_template(scenario=scenario, version=version)
        f.write(version_info)


def _build_version_template(scenario: str = "", version: str = None) -> str:
    if version is None:
        version = __version__
    return version_template.format(
        BUILD_USER=getpass.getuser(),
        BUILD_VERSION=version,
        BUILD_DATE=time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()),
        SCENARIO=scenario,
    )



"""
Browser MCP Server

This module provides MCP server functionality for browser automation and interaction.
It handles tasks such as web scraping, form submission, and automated browsing.

Main functions:
- browse_url: Opens a URL and performs specified actions
- submit_form: Fills and submits forms on web pages
"""

import json
import os
import sys
import traceback

from browser_use import Agent
from browser_use.agent.views import AgentHistoryList
from browser_use.browser.browser import Browser, BrowserConfig
from browser_use.browser.context import BrowserContext, BrowserContextConfig
from dotenv import load_dotenv
from langchain_openai import ChatOpenAI
from mcp.server.fastmcp import FastMCP
from pydantic import Field

from aworld.logs.util import logger

mcp = FastMCP("browser-server")
browser_system_prompt = """
===== NAVIGATION STRATEGY =====
1. START: Navigate to the most authoritative source for this information
   - For general queries: Use Google with specific search terms
   - For known sources: Go directly to the relevant website

2. EVALUATE: Assess each page methodically
   - Scan headings and highlighted text first
   - Look for data tables, charts, or official statistics
   - Check publication dates for timeliness

3. EXTRACT: Capture exactly what's needed
   - Take screenshots of visual evidence (charts, tables, etc.)
   - Copy precise text that answers the query
   - Note source URLs for citation

4. DOWNLOAD: Save the most relevant file to local path for further processing
   - Save the text if possible for futher text reading and analysis
   - Save the image if possible for futher image reasoning analysis
   - Save the pdf if possible for futher pdf reading and analysis

5. ROBOT DETECTION:
   - If the page is a robot detection page, abort immediately
   - Navigate to the most authoritative source for similar information instead

===== EFFICIENCY GUIDELINES =====
- Use specific search queries with key terms from the task
- Avoid getting distracted by tangential information
- If blocked by paywalls, try archive.org or similar alternatives
- Document each significant finding clearly and concisely

Your goal is to extract precisely the information needed with minimal browsing steps.
"""


@mcp.tool(description="Perform browser actions using the browser-use package.")
async def browser_use(
    task: str = Field(description="The task to perform using the browser."),
) -> str:
    """
    Perform browser actions using the browser-use package.
    Args:
        task (str): The task to perform using the browser.
    Returns:
        str: The result of the browser actions.
    """
    browser = Browser(
        config=BrowserConfig(
            headless=False,
            new_context_config=BrowserContextConfig(
                disable_security=True,
                user_agent="Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36",
                minimum_wait_page_load_time=10,
                maximum_wait_page_load_time=30,
            ),
        )
    )
    browser_context = BrowserContext(
        config=BrowserContextConfig(
            trace_path=os.getenv("LOG_FILE_PATH" + "/browser_trace.log")
        ),
        browser=browser,
    )
    agent = Agent(
        task=task,
        llm=ChatOpenAI(
            model=os.getenv("LLM_MODEL_NAME"),
            api_key=os.getenv("LLM_API_KEY"),
            base_url=os.getenv("LLM_BASE_URL"),
            model_name=os.getenv("LLM_MODEL_NAME"),
            openai_api_base=os.getenv("LLM_BASE_URL"),
            openai_api_key=os.getenv("LLM_API_KEY"),
            temperature=1.0,
        ),
        browser_context=browser_context,
        extend_system_message=browser_system_prompt,
    )
    try:
        browser_execution: AgentHistoryList = await agent.run(max_steps=50)
        if (
            browser_execution is not None
            and browser_execution.is_done()
            and browser_execution.is_successful()
        ):
            exec_trace = browser_execution.extracted_content()
            logger.info(
                ">>> 🌏 Browse Execution Succeed!\n"
                f">>> 💡 Result: {json.dumps(exec_trace, ensure_ascii=False, indent=4)}\n"
                ">>> 🌏 Browse Execution Succeed!\n"
            )
            return browser_execution.final_result()
        else:
            return f"Browser execution failed for task: {task}"
    except Exception as e:
        logger.error(f"Browser execution failed: {traceback.format_exc()}")
        return f"Browser execution failed for task: {task} due to {str(e)}"
    finally:
        await browser.close()
        logger.info("Browser Closed!")


def main():
    load_dotenv()
    print("Starting Browser MCP Server...", file=sys.stderr)
    mcp.run(transport="stdio")


# Make the module callable
def __call__():
    """
    Make the module callable for uvx.
    This function is called when the module is executed directly.
    """
    main()


sys.modules[__name__].__call__ = __call__

# Run the server when the script is executed directly
if __name__ == "__main__":
    main()



from aworld.metrics.context_manager import MetricContext
from aworld.metrics.template import MetricTemplate
from aworld.metrics.metric import MetricType

tokens_usage_histogram = MetricTemplate(
    type=MetricType.HISTOGRAM,
    name="llm_token_usage",
    unit="token",
    description="Measures number of input and output tokens used"
)

chat_choice_counter = MetricTemplate(
    type=MetricType.COUNTER,
    name="llm_generation_choice_counter",
    unit="choice",
    description="Number of choices returned by chat completions call"
)

duration_histogram = MetricTemplate(
    type=MetricType.HISTOGRAM,
    name="llm_chat_duration",
    unit="s",
    description="AI chat duration",
)

chat_exception_counter = MetricTemplate(
    type=MetricType.COUNTER,
    name="llm_chat_exception_counter",
    unit="time",
    description="Number of exceptions occurred during chat completions",
)

streaming_time_to_first_token_histogram = MetricTemplate(
    type=MetricType.HISTOGRAM,
    name="llm_streaming_time_to_first_token",
    unit="s",
    description="Time to first token in streaming chat completions",
)
streaming_time_to_generate_histogram = MetricTemplate(
    type=MetricType.HISTOGRAM,
    name="streaming_time_to_generate",
    unit="s",
    description="Time between first token and completion in streaming chat completions",
)


def record_exception_metric(exception, duration):
    '''
    record chat exception to metrics
    '''
    if MetricContext.metric_initialized():
        labels = {
            "error.type": exception.__class__.__name__,
        }
        if duration_histogram:
            MetricContext.histogram_record(
                duration_histogram, duration, labels=labels)
        if chat_exception_counter:
            MetricContext.count(
                chat_exception_counter, 1, labels=labels)


def record_streaming_time_to_first_token(duration, labels):
    '''
    Record duration of start time to first token in stream. 
    '''
    if MetricContext.metric_initialized():
        MetricContext.histogram_record(
            streaming_time_to_first_token_histogram, duration, labels=labels)


def record_streaming_time_to_generate(first_token_to_generate_duration, labels):
    '''
    Record duration the first token to response to generation
    '''
    if MetricContext.metric_initialized():
        MetricContext.histogram_record(
            streaming_time_to_generate_histogram, first_token_to_generate_duration, labels=labels)


def record_chat_response_metric(attributes,
                                prompt_tokens,
                                completion_tokens,
                                duration,
                                choices=None
                                ):
    '''
    Record chat response to metrics
    '''
    if MetricContext.metric_initialized():
        if prompt_tokens and tokens_usage_histogram:
            labels = {
                **attributes,
                "llm.prompt_usage_type": "prompt_tokens"
            }
            MetricContext.histogram_record(
                tokens_usage_histogram, prompt_tokens, labels=labels)
        if completion_tokens and tokens_usage_histogram:
            labels = {
                **attributes,
                "llm.prompt_usage_type": "completion_tokens"
            }
            MetricContext.histogram_record(
                tokens_usage_histogram, completion_tokens, labels=labels)
        if duration and duration_histogram:
            MetricContext.histogram_record(
                duration_histogram, duration, labels=attributes)
        if choices and chat_choice_counter:
            MetricContext.count(chat_choice_counter,
                                len(choices), labels=attributes)
            for choice in choices:
                if choice.get("finish_reason"):
                    finish_reason_attr = {
                        **attributes,
                        "llm.finish_reason": choice.get("finish_reason")
                    }
                    MetricContext.count(
                        chat_choice_counter, 1, labels=finish_reason_attr)



# coding: utf-8
# Copyright (c) 2025 inclusionAI.

from pydantic import BaseModel
from typing import Dict, Any, Optional, Union, List

from aworld.config import ConfigDict
from aworld.core.memory import MemoryItem

Config = Union[Dict[str, Any], ConfigDict, BaseModel]


class ActionResult(BaseModel):
    """Result of executing an action by use tool."""
    is_done: bool = False
    success: bool = False
    content: Any = None
    error: str = None
    keep: bool = False
    action_name: str = None
    tool_name: str = None
    # llm tool id
    tool_id: str = None


class Observation(BaseModel):
    """Observation information is obtained from the tools or transformed from the actions made by agents.

    It can be an agent(as a tool) in the swarm or a tool in the virtual environment.
    """
    # default is None, means the main virtual environment or swarm
    container_id: Optional[str] = None
    # Observer who obtains observation, default is None for compatible, means an agent name or a tool name
    observer: Optional[str] = None
    # default is None for compatible, means with its action/ability name of an agent or a tool
    # NOTE: The only ability of an agent as a tool is handoffs
    ability: Optional[str] = None
    # The agent wants the observation to be created, default is None for compatible.
    from_agent_name: Optional[str] = None
    # To which agent should the observation be given, default is None for compatible.
    to_agent_name: Optional[str] = None
    # general info for agent
    content: Optional[Any] = None
    # dom_tree is a str or DomTree object
    dom_tree: Optional[Union[str, Any]] = None
    image: Optional[str] = None  # base64
    action_result: Optional[List[ActionResult]] = []
    # for video or image list
    images: Optional[List[str]] = []
    # extend key value pair. `done` is an internal key
    info: Optional[Dict[str, Any]] = {}


class StatefulObservation(Observation):
    """Observations with contextual states."""
    context: List[MemoryItem]


class ParamInfo(BaseModel):
    name: str | None = None
    type: str = "str"
    required: bool = False
    desc: str = None
    default_value: Any = None


class ToolActionInfo(BaseModel):
    name: str
    input_params: Dict[str, ParamInfo] = {}
    desc: str = None


class ActionModel(BaseModel):
    tool_name: Optional[str] = None
    tool_id: Optional[str] = None
    # agent name
    agent_name: Optional[str] = None
    # action_name is a tool action name by agent policy.
    action_name: Optional[str] = None
    params: Optional[Dict[str, Any]] = {}
    policy_info: Optional[Any] = None


class TaskItem(BaseModel):
    data: Any
    msg: str = None
    stop: bool = False
    success: bool = False
    action_name: Optional[str] = None
    params: Optional[Dict[str, Any]] = {}
    policy_info: Optional[Any] = None



from typing import Any, Callable
from .asgi import TraceMiddleware
from aworld.trace.instrumentation import Instrumentor
from aworld.trace.base import TraceProvider, get_tracer_provider
from aworld.trace.instrumentation.http_util import (
    get_excluded_urls,
    parse_excluded_urls,
)
from aworld.utils.import_package import import_packages
from aworld.logs.util import logger

import_packages(['fastapi'])  # noqa
import fastapi  # noqa


class _InstrumentedFastAPI(fastapi.FastAPI):
    """Instrumented FastAPI class."""
    _tracer_provider: TraceProvider = None
    _excluded_urls: list[str] = None
    _server_request_hook: Callable = None
    _client_request_hook: Callable = None
    _client_response_hook: Callable = None
    _instrumented_fastapi_apps = set()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        tracer = self._tracer_provider.get_tracer(
            "aworld.trace.instrumentation.fastapi")

        self.add_middleware(
            TraceMiddleware,
            tracer=tracer,
            excluded_urls=self._excluded_urls,
            server_request_hook=self._server_request_hook,
            client_request_hook=self._client_request_hook,
            client_response_hook=self._client_response_hook
        )

        self._is_instrumented_by_trace = True
        self._instrumented_fastapi_apps.add(self)

    def __del__(self):
        if self in self._instrumented_fastapi_apps:
            self._instrumented_fastapi_apps.remove(self)


class FastAPIInstrumentor(Instrumentor):
    """FastAPI Instrumentor."""
    _original_fastapi = None

    @staticmethod
    def uninstrument_app(app: fastapi.FastAPI):
        app.user_middleware = [
            x
            for x in app.user_middleware
            if x.cls is not TraceMiddleware
        ]
        app.middleware_stack = app.build_middleware_stack()
        app._is_instrumented_by_trace = False

    def instrumentation_dependencies(self) -> dict[str, Any]:
        return {"fastapi": fastapi}

    def _instrument(self, **kwargs):
        self._original_fastapi = fastapi.FastAPI
        _InstrumentedFastAPI._tracer_provider = kwargs.get("tracer_provider")
        _InstrumentedFastAPI._server_request_hook = kwargs.get(
            "server_request_hook"
        )
        _InstrumentedFastAPI._client_request_hook = kwargs.get(
            "client_request_hook"
        )
        _InstrumentedFastAPI._client_response_hook = kwargs.get(
            "client_response_hook"
        )
        excluded_urls = kwargs.get("excluded_urls")
        _InstrumentedFastAPI._excluded_urls = (
            get_excluded_urls("FASTAPI")
            if excluded_urls is None
            else parse_excluded_urls(excluded_urls)
        )
        fastapi.FastAPI = _InstrumentedFastAPI

    def _uninstrument(self, **kwargs):
        for app in _InstrumentedFastAPI._instrumented_fastapi_apps:
            self.uninstrument_app(app)
        _InstrumentedFastAPI._instrumented_fastapi_apps.clear()
        fastapi.FastAPI = self._original_fastapi


def instrument_fastapi(excluded_urls: str = None,
                       server_request_hook: Callable = None,
                       client_request_hook: Callable = None,
                       client_response_hook: Callable = None,
                       tracer_provider: TraceProvider = None,
                       **kwargs: Any,
                       ):
    kwargs.update({
        "excluded_urls": excluded_urls,
        "server_request_hook": server_request_hook,
        "client_request_hook": client_request_hook,
        "client_response_hook": client_response_hook,
        "tracer_provider": tracer_provider or get_tracer_provider(),
    })
    FastAPIInstrumentor().instrument(**kwargs)
    logger.info("FastAPI instrumented.")



# coding: utf-8
# Copyright (c) 2025 inclusionAI.
import time
import uuid
from typing import List

from pydantic import BaseModel


class Session(BaseModel):
    session_id: str = str(uuid.uuid1().hex)
    last_update_time: float = time.time()
    trajectories: List = []



# coding: utf-8
# Copyright (c) 2025 inclusionAI.
import json

import numpy as np


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super().default(obj)
